#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

from __future__ import division

import math
import os
import random

from pyomo.common.dependencies import yaml, yaml_load_args
import pyomo.common.plugin
from pyomo.pysp import phextension
from pyomo.pysp.phutils import *
from pyomo.pysp.phsolverserverutils import \
    transmit_external_function_invocation_to_worker
from pyomo.pysp.generators import \
    scenario_tree_node_variables_generator_noinstances
from pyomo.core.base import *

import six
from six import iteritems
from six.moves import xrange

###############
def _parse_yaml_file(ph, filename):

    with open(filename) as f:
        config_data = yaml.load(f, **yaml_load_args)

    for node_or_stage_name, variable_dicts in iteritems(config_data):

        # find the node or stage object in the scenario tree
        # and convert it to a list of nodes
        node_list = None
        for stage in ph._scenario_tree._stages[:-1]:
            if node_or_stage_name == stage._name:
                node_list = stage._tree_nodes
                break
            else:
                for tree_node in stage._tree_nodes:
                    if tree_node._name == node_or_stage_name:
                        node_list = [tree_node]
                        break
                if node_list is not None:
                    break

        if len(node_list) == 0:
            raise RuntimeError("***WW PH Extension: %s does not name an "
                               "existing non-leaf Stage or Node in the "
                               "scenario tree" % node_or_stage_name)

        for variable_string, variable_config_data in iteritems(variable_dicts):

            variable_name = None
            index_template = None
            if isVariableNameIndexed(variable_string):
                variable_name, index_template = \
                    extractVariableNameAndIndex(variable_string)
            else:
                variable_name = variable_string
                index_template = None

            match_variable_ids = {}
            for node in node_list:
                node_match_variable_ids = match_variable_ids[node._name] = []
                for variable_id, (name,index) in iteritems(node._variable_ids):
                    if (variable_name == name):
                        if (index_template is None) and (index != None):
                            raise RuntimeError("***WW PH Extension: Illegal "
                                               "match template="
                                               +variable_string+" specified "
                                               "in file="+filename)
                        if indexMatchesTemplate(index, index_template):
                            node_match_variable_ids.append(variable_id)
            # we shouldn't have any duplicates or empties
            for node in node_list:
                assert len(set(match_variable_ids)) == len(match_variable_ids)
                if len(match_variable_ids[node._name]) == 0:
                    raise RuntimeError("No variables matching string '%s' "
                                       "were found in scenario tree node %s"
                                       % (variable_string, node._name))

            for config_name, config_value in iteritems(variable_config_data):

                yield config_name, config_value, match_variable_ids

####################

#
# Create a dictionary mapping scenario tree id to
# variable bounds
#
def collect_node_variable_bounds(tree_node):
    assert tree_node._variable_datas is not None
    var_bounds = {}
    for variable_id, vardatas in iteritems(tree_node._variable_datas):
        vardata = vardatas[0][0]
        if not vardata.is_expression_type():
            var_bounds[variable_id] = (vardata.lb, vardata.ub)
        else:
            # Expression
            var_bounds[variable_id] = (None,None)

    return var_bounds

# This function will be executed on a phsolverserver using
# the transmit_external_function_invocation_to_worker utility
def external_collect_variable_bounds(ph,
                                     scenario_tree,
                                     scenario_or_bundle,
                                     node_list):

    variable_bounds = {}
    for node_name in node_list:
        tree_node = scenario_tree._tree_node_map[node_name]
        variable_bounds[node_name] = collect_node_variable_bounds(tree_node)

    return variable_bounds

#==================================================
class wwphextension(pyomo.common.plugin.SingletonPlugin):

    pyomo.common.plugin.implements(phextension.IPHExtension)

    # the below is a hack to get this extension into the
    # set of IPHExtension objects, so it can be queried
    # automagically by PH.
    pyomo.common.plugin.alias("WWPHExtension")

    def __init__(self, *args, **kwds):

        # attempt to handle cases where scripting people
        # skip ph iteration 0
        self._iteration_0_called = False

        # TBD - migrate all of the self attributes defined on-the-fly
        #       in the post-post-initialization routine here!
        self._valid_suffix_names = [\
            'Iter0FixIfConvergedAtLB',   # has default
            'Iter0FixIfConvergedAtUB',   # has default
            'Iter0FixIfConvergedAtNB',   # has default
            'CostForRho',
            'FixWhenItersConvergedAtLB', # has default
            'FixWhenItersConvergedAtUB', # has default
            'FixWhenItersConvergedAtNB', # has default
            'CanSlamToLB',               # has default
            'CanSlamToMin',              # has default
            'CanSlamToUB',               # has default
            'CanSlamToMax',              # has default
            'CanSlamToAnywhere',         # has default
            'SlammingPriority']
        self._AnnotationTypes = {}
        # Real
        self._AnnotationTypes['going_price'] = None
        # String
        self._AnnotationTypes['obj_effect_family_name'] = None
        # Real
        self._AnnotationTypes['obj_effect_family_factor'] = None
        # non-negative int
        self._AnnotationTypes['decision_hierarchy_level'] = None
        self._AnnotationTypes['feasibility_direction'] = \
            ['down', 'up', 'either', 'None']
        # int
        self._AnnotationTypes['relax_int'] = None
        # int
        self._AnnotationTypes['reasonable_int'] = None
        # int
        self._AnnotationTypes['low_int'] = None

        assert len(set(self._valid_suffix_names).\
                   intersection(self._AnnotationTypes.keys())) == 0

        self._suffixes = {}
        for name in self._valid_suffix_names:
            self._suffixes[name] = {}
        for name in self._AnnotationTypes:
            self._suffixes[name] = {}

        # formerly known as self._suffixes["my_stage"]
        self._variable_stage = {}

        self._configuration_filename = None
        self._suffix_filename = None
        self._annotation_filename = None

        # we track various actions performed by this extension following a
        # PH iteration - useful for reporting purposes, and for other plugins
        # to react to.
        self.variables_with_detected_cycles = []
        self.variable_fixed = False


#==================================================

    def process_suffix_file(self, ph):

        self.slam_list = []

        if self._suffix_filename is None:
            return

        if os.path.exists(self._suffix_filename) is False:
            raise RuntimeError("***WW PH Extension: The suffix "
                               "file "+self._suffix_filename+
                               " either does not exist or cannot be read")

        print("WW PH Extension: Loading variable suffixes "
              "from file="+self._suffix_filename)

        for suffix_name, suffix_value, variable_ids in \
              _parse_yaml_file(ph, self._suffix_filename):

            if suffix_name not in self._valid_suffix_names:
                raise RuntimeError("***WW PH Extension: %s is not a valid "
                                   "suffix name" % (suffix_name))

            # decide what type of suffix value we're dealing with.
            is_int = False
            is_bool = False
            converted_value = None
            try:
                converted_value = bool(suffix_value)
                is_bool = True
            except ValueError:
                pass
            try:
                converted_value = int(suffix_value)
                is_int = True
            except ValueError:
                pass

            if (not is_int) and (not is_bool):
                raise RuntimeError("WW PH Extension unable to deduce type of data "
                                   "referenced in ww ph extension suffix file"
                                   "="+self._suffix_filename+"; value="
                                   +suffix_value)

            self_suffix_dict = self._suffixes[suffix_name]
            for node_name, node_variable_ids in iteritems(variable_ids):
                for variable_id in node_variable_ids:
                    self_suffix_dict[variable_id] = converted_value

        self.slam_list = list(self._suffixes["SlammingPriority"].keys())

        if self.slam_list != []:
            #
            # We have been ordering this list largest slamming
            # priority to smallest (meaning a larger value for
            # slamming priority implies we're first in line to get
            # fixed). I can't decide if that is the most natural way
            # to interpret this, but I'm leaving it as it will likely
            # change solutions if changed.
            #
            self.slam_list.sort(key=self._suffixes["SlammingPriority"].get,
                                reverse=True)

    def process_annotation_file(self, ph):
        # note: these suffixes can have a string value very similar to
        # suffixes, except for the following: annotation names are
        # from a restricted list, and annotations can have various
        # types that might depend on the name of the annotation
        # not all type checking will be done here, but some might be

        self._obj_family_normalized_rho = {}

        if self._annotation_filename is None:
            return

        if os.path.exists(self._annotation_filename) is False:
            raise RuntimeError("***WW PH Extension: The annotation file "
                               +self._annotation_filename+" either does "
                               "not exist or cannot be read")

        print("WW PH Extension: Loading variable annotations from "
              "file="+self._annotation_filename)

        for annotation_name, annotation_value, variable_ids in \
              _parse_yaml_file(ph, self._annotation_filename):

            # check for some input errors
            if annotation_name not in self._AnnotationTypes:
                print("Error encountered.")
                print("Here are the annotations that can be given "
                      "(they are case sensitive):")
                for i in self._AnnotationTypes:
                    print(i)
                raise RuntimeError("WW ph extension annotation file="
                                   +self._annotation_filename+"; "
                                   "contains unknown annotation: "+annotation_name)

            # check for some more input errors
            if self._AnnotationTypes[annotation_name] is not None:
                if annotation_value not in self._AnnotationTypes[annotation_name]:
                    raise RuntimeError("WW ph extension annotation file="
                                       +self._annotation_filename+"; "
                                       "contains unknown annotation value="
                                       +annotation_value+" for: "+annotation_name)

            # if this is a new obj effect family, then we need new maps
            if annotation_name == 'obj_effect_family_name':
                if annotation_value not in self._obj_family_normalized_rho:
                    self._obj_family_normalized_rho[annotation_value] = 0.0

            self_suffix_dict = self._suffixes[annotation_name]
            for node_name, node_variable_ids in iteritems(variable_ids):
                for variable_id in node_variable_ids:
                    self_suffix_dict[variable_id] = converted_value

    #
    # Store the variable bounds on the tree node by variable id
    # If this is parallel ph, then we need to do this by transmitting
    # an external function invocation that collects the variable
    # bounds from a given list of node names. We do this in such
    # a way as to minimize the amount of data we need to transfer
    # (e.g., in the case of a two stage problem, we request information
    # about the root node from only one of the PH solver servers)
    #

    def _collect_variable_bounds(self,ph):

        if not isinstance(ph._solver_manager,
                          pyomo.solvers.plugins.smanager.phpyro.SolverManager_PHPyro):

            for stage in ph._scenario_tree._stages[:-1]:

                for tree_node in stage._tree_nodes:

                    tree_node._variable_bounds = \
                        collect_node_variable_bounds(tree_node)

        else:

            nodes_needed = {}
            if ph._scenario_tree.contains_bundles():
                for tree_node in ph._scenario_tree._tree_nodes:
                    if not tree_node.is_leaf_node():
                        representative_scenario = tree_node._scenarios[0]._name
                        representative_bundle = None
                        for bundle in ph._scenario_tree._scenario_bundles:
                            if representative_scenario in bundle._scenario_names:
                                representative_bundle = bundle._name
                                break
                        assert representative_bundle != None
                        if representative_bundle not in nodes_needed:
                            nodes_needed[representative_bundle] = \
                                [tree_node._name]
                        else:
                            nodes_needed[representative_bundle].\
                                append(tree_node._name)
            else:
                for tree_node in ph._scenario_tree._tree_nodes:
                    if not tree_node.is_leaf_node():
                        representative_scenario = tree_node._scenarios[0]._name
                        if representative_scenario not in nodes_needed:
                            nodes_needed[representative_scenario] = \
                                [tree_node._name]
                        else:
                            nodes_needed[representative_scenario].\
                                append(tree_node._name)

            action_handle_map = {}
            for object_name, node_list in iteritems(nodes_needed):
                function_args = (node_list,)
                new_action_handle = \
                    transmit_external_function_invocation_to_worker(
                        ph,
                        object_name,
                        "pyomo.pysp.plugins.wwphextension",
                        "external_collect_variable_bounds",
                        return_action_handle=True,
                        function_args=function_args)

                action_handle_map[new_action_handle] = object_name

            num_results_so_far = 0
            num_results = len(action_handle_map)
            while (num_results_so_far < num_results):
                ah = ph._solver_manager.wait_any()
                results = ph._solver_manager.get_results(ah)
                object_name = action_handle_map[ah]
                for node_name, node_bounds in iteritems(results):
                    tree_node = ph._scenario_tree._tree_node_map[node_name]
                    tree_node._variable_bounds = node_bounds
                num_results_so_far += 1

    def reset(self, ph):
        self.__init__()

    def pre_ph_initialization(self,ph):
        # we don't need to intefere with PH initialization.
        pass

    def post_instance_creation(self, ph):
        # we don't muck with the instances.
        pass

    def post_ph_initialization(self, ph):

        # Be verbose (even if ph is not)
        self.Verbose = False

        # set up "global" record keeping.
        self.cumulative_discrete_fixed_count = 0
        self.cumulative_continuous_fixed_count = 0

        # we always track convergence of continuous variables, but we
        # may not want to fix/slam them.
        self.fix_continuous_variables = False

        # there are occasions where we might want to fix any values at
        # the end of the run if the scenarios agree - even if the
        # normal fixing criterion (e.g., converged for N iterations)
        # don't apply. one example is when the term-diff is 0, at
        # which point you really do have a solution. currently only
        # applies to discrete variables.
        self.fix_converged_discrete_variables_at_exit = False

        # set up the mipgap parameters (zero means ignore)
        # note:
        # because we lag one iteration, the final will be less than
        # requested initial and final refer to PH iteration 1 and PH
        # iteration X, where X is the iteration at which the
        # convergence metric hits 0.
        self.Iteration0MipGap = 0.0
        self.InitialMipGap = 0.0
        self.FinalMipGap = 0.0
        self.MipGapExponent = 1.0 # linear by default

        # the comparison tolerance for two hashed weight values.
        # by default, 1e-5 should suffice. however, depending on
        # solution structure and mipgap, non-default values may
        # be necessary.
        # IMPT (and new): this value is interpreted as a percentage,
        # i.e., a value of 1 indicates a 1% difference in hash values
        # is sufficient to trigger a cycle detection claim.
        self.WHashComparisonTolerance = 1e-5

        # it is occasionally useful to see the W hash histories
        # when a cycle is detected, to examine patterns in support
        # of identifying an appropriate comparison tolerance.
        # defaults to False.
        self.ReportCycleWHashDetail = False

        # use the convergence metric value from iteration 0 as the
        # basis for scaling the mipgap in subsequent iterations. if
        # set to False, we instaed start scaling from the convergence
        # metric value following iteration 1. the latter arguably
        # makes more sense, as it accounts for weights and such, i.e.,
        # some sub-prolem interactions.
        self.StartMipGapFromIter0Metric = True

        # when cylcing is detected for a variable, the
        # rho will be multiplied by this factor.
        # NOTE: 1 is a no-op
        self.RhoReductionFactor = 1.0

        # when cycling is detected for a variable, the
        # rho on a per scenario basis will be perturbed
        # by the following factor (uniformly sampled,
        # plus/minus this factor).
        # NOTE: 0 is a no-op
        self.ScenarioRhoPerturbation = 0.0

        # "Read" the defaults for parameters that control fixing
        # (these defaults can be overridden at the variable level) for
        # all six of these, zero means don't do it.
        self.Iter0FixIfConvergedAtLB = 0 # 1 or 0
        self.Iter0FixIfConvergedAtUB = 0  # 1 or 0
        self.Iter0FixIfConvergedAtNB = 0  # 1 or 0 (converged to a non-bound)
        # TBD: check for range errors for all six of these
        self.FixWhenItersConvergedAtLB = 10
        self.FixWhenItersConvergedAtUB = 10
        # converged to a non-bound
        self.FixWhenItersConvergedAtNB = 12
        self.FixWhenItersConvergedContinuous = 0

        # "default" slamming parms:
        self.CanSlamToLB = False
        self.CanSlamToMin = False
        self.CanSlamToAnywhere = True
        self.CanSlamToMax = False
        self.CanSlamToUB = False
        # zero means "slam at will"
        self.PH_Iters_Between_Cycle_Slams = 1
        # this is a simple heuristic - and probably a bad one!
        self.SlamAfterIter = len(ph._scenario_tree._stages[-1]._tree_nodes)
        # slam only if the convergence metric is not improving. defaults to
        # false. in general, if the convergence metric is improving, then
        # we would like PH to enforce non-antipicativity on its own. that
        # said, there are counter-examples - particularly when we have
        # continuous variables.
        self.SlamOnlyIfNotConverging = False

        # default params associated with fixing due to weight vector oscillation.
        self.CheckWeightOscillationAfterIter = 0
        self.FixIfWeightOscillationCycleLessThan = 10

        # flags enabling various rho computation schemes.
        self.ComputeRhosWithpreSEP = False
        self.ComputeRhosWithSEP = False
        self.RhosSEPMult = None  # the user needs an error if not set...

        self.CycleLengthSlamThreshold = \
            len(ph._scenario_tree._stages[-1]._tree_nodes)
        self.W_hash_history_len = max(100, self.CycleLengthSlamThreshold+1)

        # do I report potential cycles, i.e., those too short to base
        # slamming on?
        self.ReportPotentialCycles = False

        # end of parms

        self._last_slam_iter = -1    # dynamic

        # constants for W vector hashing (low cost rand() is all we need)
        # note: July 09, dlw is planning to eschew pre-computed random
        #   vector (dlw April 2014: "huh?")
        # another note: the top level reset is OK, but really it should
        #   done way down at the index level with a serial number (stored)
        #   to avoid correlated hash errors
        # the hash seed was deleted in 1.1 and we seed with the

        # we will reset for dynamic rand vector generation
        self.W_hash_seed = 17
        # current rand
        self.W_hash_rand_val = self.W_hash_seed
        # a,b, and c are for a simple generator
        self.W_hash_a = 1317
        self.W_hash_b = 27699
        # that period should be usually long enough for us!
        # (assuming fewer than c scenarios)
        self.W_hash_c = 131072

        my_stage_suffix = self._variable_stage

        # set up tree storage for statistics that we care about tracking.
        for stage in ph._scenario_tree._stages[:-1]:

            for tree_node in stage._tree_nodes:

                # we're adding a lot of statistics / tracking data to
                # each tree node.  these are all maps from variable
                # name to a parameter that tracks the corresponding
                # information.
                tree_node._num_iters_converged = \
                    dict.fromkeys(tree_node._variable_ids,0)
                tree_node._last_converged_val = \
                    dict.fromkeys(tree_node._variable_ids,0.5)
                tree_node._w_hash = \
                    dict(((variable_id,ph_iter),0) \
                         for variable_id in tree_node._variable_ids \
                         for ph_iter in ph._iteration_index_set)
                # sign vector for weights at the last PH iteration
                tree_node._w_sign_vector = \
                    dict.fromkeys(tree_node._variable_ids,[])
                # the number of iterations since the last flip in the
                # sign (TBD - of any scenario in the vector)?
                tree_node._w_last_sign_flip_iter = \
                    dict.fromkeys(tree_node._variable_ids,0)

                my_stage_suffix.update((variable_id,stage) \
                                       for variable_id in tree_node._variable_ids)

        if self._configuration_filename is not None:
            if os.path.exists(self._configuration_filename) is True:
                print("WW PH Extension: Loading user-specified configuration "
                      "from file=" + self._configuration_filename)
                try:
                    if six.PY3:
                        with open(self._configuration_filename,"rb") as f:
                            code = compile(f.read(), self._configuration_filename, 'exec')
                            exec(code) in globals(), locals()
                    else:
                        execfile(self._configuration_filename)
                except:
                    raise RuntimeError("Failed to load WW PH configuration file="
                                       +self._configuration_filename)
            else:
                raise RuntimeError("***WW PH Extension: The configuration file "
                                   +self._configuration_filename+" either does "
                                   "not exist or cannot be read")
        else:
            print("WW PH Extension: No user-specified configuration file "
                  "supplied - using defaults")

        # process any suffix and/or annotation data, if they exists.
        self.process_suffix_file(ph)
        self.process_annotation_file(ph)

        # set up the mip gap for iteration 0.
        if self.Iteration0MipGap > 0.0:
            print("WW PH Extension: Setting mipgap to %.7f, as specified in "
                  "the configuration file" % self.Iteration0MipGap)
            ph._mipgap = self.Iteration0MipGap

        # search for the presence of the normalized-term-diff converger
        # (first priority) or the standard term-diff converger (second priority).
        self.converger_to_use = None
        for converger in ph._convergers:
            if converger._name == "Normalized term diff":
                self.converger_to_use = converger
                break
            elif converger._name == "Term diff":
                self.converger_to_use = converger
                break
        if self.converger_to_use == None:
            raise RuntimeError("WW PH extension plugin failed to locate a normalized term-diff or term-diff converger on the PH object")

#==================================================
    def post_iteration_0_solves(self, ph):

        self._iteration_0_called = True

        self._collect_variable_bounds(ph)

        # Collect suffix dictionaries that we will check inside the
        # next loop
        Iter0FixIfConvergedAtLB = self._suffixes['Iter0FixIfConvergedAtLB']
        Iter0FixIfConvergedAtUB = self._suffixes['Iter0FixIfConvergedAtUB']
        Iter0FixIfConvergedAtNB = self._suffixes['Iter0FixIfConvergedAtNB']
        CostForRho = self._suffixes['CostForRho']

        for stage, tree_node, variable_id, variable_values, is_fixed, is_stale in \
              scenario_tree_node_variables_generator_noinstances(
                  ph._scenario_tree,
                  includeDerivedVariables=False,
                  includeLastStage=False,
                  sort=True):

            if (is_stale is False):
                """ preSEP is not a good idea; E[W] will not be 0
                if (self.ComputeRhosWithpreSEP) and (variable_id in CostForRho):
                    # Watson and Woodruff Comp. Mgt. Sci.
                    # right before the SEP expressions: a per-scenario rho

                    node_average = tree_node._averages[variable_id]
                    for scenario in tree_node._scenarios:
                        var_value = scenario._x[tree_node._name][variable_id]
                        # dlw: not sure that max(.,1) is the right thing...
                        rhoval = CostForRho[variable_id] * self.RhosSEPMult / max(math.fabs(var_value - node_average), 1)
                        if self.Verbose or ph._verbose:
                            print ("wwphextension is setting rho to %f for scenario=%s, variable id=%s" % (rhoval, str(scenario) , str(variable_id)))
                    ph.setRhoOneScenario(
                        tree_node,
                        scenario,
                        variable_id,
                        rhoval)
                """
                if (self.ComputeRhosWithSEP) and (variable_id in CostForRho):

                    node_average = tree_node._averages[variable_id]
                    deviation_from_average = 0.0
                    sumprob = 0.0
                    for scenario in tree_node._scenarios:
                        scenario_probability = scenario._probability
                        var_value = scenario._x[tree_node._name][variable_id]
                        deviation_from_average += \
                            (scenario_probability * \
                             math.fabs(var_value - node_average))
                        sumprob += scenario_probability
                    # expected deviation
                    deviation_from_average /= sumprob

                    node_min = \
                        self.Int_If_Close_Enough(ph,
                                                 tree_node._minimums[variable_id])
                    node_max = \
                        self.Int_If_Close_Enough(ph,
                                                 tree_node._maximums[variable_id])

                    if tree_node.is_variable_discrete(variable_id):
                        denominator = max(node_max - node_min + 1, 1)
                    else:
                        denominator = max(deviation_from_average, 1)

                    # CostForRho are the costs to be used as the
                    # numerator in the rho computation below.
                    if self.Verbose or ph._verbose:
                        print ("wwphextension is setting rho to %f for variable id=%s" % (CostForRho[variable_id] * self.RhosSEPMult / denominator , str(variable_id)))
                    ph.setRhoAllScenarios(
                        tree_node,
                        variable_id,
                        CostForRho[variable_id] * self.RhosSEPMult / denominator)

                if is_fixed is False:

                    if tree_node.is_variable_discrete(variable_id):
                        node_min = \
                            self.Int_If_Close_Enough(
                                ph,
                                tree_node._minimums[variable_id])
                        node_max = \
                            self.Int_If_Close_Enough(
                                ph,
                                tree_node._maximums[variable_id])

                        # update convergence prior to checking for fixing.
                        self._int_convergence_tracking(ph,
                                                       tree_node,
                                                       variable_id,
                                                       node_min,
                                                       node_max)

                        lb = self.Iter0FixIfConvergedAtLB
                        if variable_id in Iter0FixIfConvergedAtLB:
                            lb = Iter0FixIfConvergedAtLB[variable_id]

                        ub = self.Iter0FixIfConvergedAtUB
                        if variable_id in Iter0FixIfConvergedAtUB:
                            ub = Iter0FixIfConvergedAtUB[variable_id]

                        nb = self.Iter0FixIfConvergedAtNB
                        if variable_id in Iter0FixIfConvergedAtNB:
                            nb = Iter0FixIfConvergedAtNB[variable_id]

                        if self._should_fix_discrete_due_to_conv(tree_node,
                                                                 variable_id,
                                                                 lb,
                                                                 ub,
                                                                 nb):
                            self._fix_var(ph, tree_node, variable_id, node_min)
                        elif self.W_hash_history_len > 0:
                            # if not fixed, then hash - no slamming at iteration 0

                            # obviously not checking for cycles at iteration 0!
                            self._w_history_accounting(ph, tree_node, variable_id)

                    else:

                        node_min = tree_node._minimums[variable_id]
                        node_max = tree_node._maximums[variable_id]

                        self._continuous_convergence_tracking(ph,
                                                              tree_node,
                                                              variable_id,
                                                              node_min,
                                                              node_max)

# jpw: not sure if we care about cycle detection in continuous variables?
#                           if self.W_hash_history_len > 0:
#                              self._w_history_accounting(ph, tree_node, variable_id)


#==================================================
    def post_iteration_0(self, ph):

        pass

#==================================================

    def pre_iteration_k_solves(self, ph):

        if (ph._current_iteration == 1) and \
           (not self._iteration_0_called):
            print("WW PH extension detected iteration 0 solves were skipped. "
                  "Executing post_iteration_0_solves plugin code.")
            self.post_iteration_0_solves(ph)
            self._iteration_0_called = True

        # NOTE: iteration 0 mipgaps are handled during extension
        # initialization, following PH initialization.

        if (self.InitialMipGap > 0 and self.FinalMipGap >= 0) and \
           self.InitialMipGap > self.FinalMipGap:

            if ph._current_iteration == 1:
                print("WW PH Extension: Setting mipgap to %.7f, as "
                      "specified in the configuration file" % self.InitialMipGap)
                ph._mipgap = self.InitialMipGap

            else:

                m0 = self.converger_to_use._metric_history[0]
                m1 = self.converger_to_use._metric_history[1]
                m = self.converger_to_use.lastMetric()
                mlast = self.converger_to_use._convergence_threshold

                basis_value = m0
                if self.StartMipGapFromIter0Metric == False:
                    basis_value = m1

                x = 1.0 - (m-mlast)/(basis_value-mlast)
                y = 1.0 - x**self.MipGapExponent

                g0 = self.InitialMipGap
                glast = self.FinalMipGap

                new_gap = y * (g0-glast) + glast

                if new_gap > g0:
                    print("WW PH Extension: ***CAUTION - setting mipgap to "
                          "thresholded maximal initial value=%.7f; "
                          "unthresholded value=%.7f" % (g0, new_gap))
                    new_gap = g0
                else:
                    print("WW PH Extension: Setting mipgap to %.7f, based "
                          "on current value of the convergence metric"
                          % (new_gap))
                ph._mipgap = new_gap

#==================================================
    def post_iteration_k_solves(self, ph):

        # Collect suffix dictionaries that we will check inside the
        # next loop
        FixWhenItersConvergedAtLB = self._suffixes['FixWhenItersConvergedAtLB']
        FixWhenItersConvergedAtUB = self._suffixes['FixWhenItersConvergedAtUB']
        FixWhenItersConvergedAtNB = self._suffixes['FixWhenItersConvergedAtNB']

        # track all variables for which cycles have been detected -
        # other plugins may want to react to / report on this
        # information.
        self.variables_with_detected_cycles = [] # consists of tuples of variable id, name, cycle length, cycle message, and tree node
        self.variable_fixed = False

        print ("WW PH Extension: Analyzing variables for fixing")

        for stage, tree_node, variable_id, variable_values, is_fixed, is_stale in \
              scenario_tree_node_variables_generator_noinstances(
                  ph._scenario_tree,
                  includeDerivedVariables=False,
                  includeLastStage=False,
                  sort=True):

            (lbval, ubval) = tree_node._variable_bounds[variable_id]
            # if the variable is stale, don't waste time fixing and
            # cycle checking. for one, the code will crash :-) due to
            # None values observed during the cycle checking
            # computation.
            if (is_stale is False) and (is_fixed is False) \
                and (lbval is None or ubval is None or lbval != ubval):

                variable_name, index = tree_node._variable_ids[variable_id]
                full_variable_name = variable_name+indexToString(index)
                if tree_node.is_variable_discrete(variable_id):
                    node_min = \
                        self.Int_If_Close_Enough(ph,
                                                 tree_node._minimums[variable_id])
                    node_max = \
                        self.Int_If_Close_Enough(ph,
                                                 tree_node._maximums[variable_id])

                    # update convergence prior to checking for fixing.
                    self._int_convergence_tracking(ph,
                                                   tree_node,
                                                   variable_id,
                                                   node_min,
                                                   node_max)

                    lb = self.FixWhenItersConvergedAtLB
                    if variable_id in FixWhenItersConvergedAtLB:
                        lb = FixWhenItersConvergedAtLB[variable_id]

                    ub = self.FixWhenItersConvergedAtUB
                    if variable_id in FixWhenItersConvergedAtUB:
                        ub = FixWhenItersConvergedAtUB[variable_id]

                    nb = self.FixWhenItersConvergedAtNB
                    if variable_id in FixWhenItersConvergedAtNB:
                        nb = FixWhenItersConvergedAtNB[variable_id]

                    if self._should_fix_discrete_due_to_conv(tree_node,
                                                             variable_id,
                                                             lb,
                                                             ub,
                                                             nb):

                        self._fix_var(ph, tree_node, variable_id, node_min)

                    else:

                        # check to see if a cycle exists for this variable,
                        # and record it if it does - cycle breaking will
                        # happen later, if it happens at all.
                        if self.W_hash_history_len > 0:

                            self._w_history_accounting(ph, tree_node, variable_id)

                            computed_cycle_length, msg = \
                                self.compute_cycle_length(ph,
                                                          tree_node,
                                                          variable_id,
                                                          False)

                            if computed_cycle_length > 0:
                                if (computed_cycle_length >= self.CycleLengthSlamThreshold) or \
                                        self.ReportPotentialCycles:
                                    self.variables_with_detected_cycles.append(
                                        (variable_id, full_variable_name,
                                         computed_cycle_length, msg, tree_node))

                else:

                    # obviously don't round in the continuous case.
                    node_min = tree_node._minimums[variable_id]
                    node_max = tree_node._maximums[variable_id]

                    # update w statistics for whatever nefarious
                    # purposes are enabled.
                    if self.W_hash_history_len > 0:
                        self._w_history_accounting(ph,
                                                   tree_node,
                                                   variable_id)

                        # update convergence prior to checking for
                        # fixing.
                        self._continuous_convergence_tracking(ph,
                                                              tree_node,
                                                              variable_id,
                                                              node_min,
                                                              node_max)

                    if self._should_fix_continuous_due_to_conv(tree_node,
                                                               variable_id):
                        # fixing to max value for safety (could only
                        # be an issue due to tolerances).
                        self._fix_var(ph, tree_node, variable_id, node_max)
                        # note: we currently don't slam continuous variables!

        # while cycles could be broken by variable fixing, the
        # notions of cycle breaking and variable fixing are really
        # independent. in particular, cycles is indicative of bad
        # rhos and/or xbar estimates, indicating the weights
        # probably need to be adjusted in any case.
        if len(self.variables_with_detected_cycles) > 0:
            print ("WW PH Extension: Cycles were detected - initiating cycle analysis and remediation")
            if self.Verbose or ph._verbose:
                    print ("WW PH Extension: Cycle information:")
                    print ("Variable            Cycle Length     Cycle Details")
            for variable_id, full_variable_name, computed_cycle_length, cycle_msg, tree_node in self.variables_with_detected_cycles:
                if self.Verbose or ph._verbose:
                    print ("%20s %4d         %s" % (full_variable_name, computed_cycle_length, cycle_msg))
                if (computed_cycle_length \
                       >= self.CycleLengthSlamThreshold) and \
                       ((ph._current_iteration - self._last_slam_iter) \
                             > self.PH_Iters_Between_Cycle_Slams):
                    # TBD: we may not want to slam
                    # immediately - it may disappear on
                    # its own after a few iterations,
                    # depending on what other variables do.
                    # note: we are *not* slamming the
                    # offending variable, but a selected variable
                    if self.Verbose or ph._verbose:
                        print(cycle_msg)
                        print("        Cycle length exceeds iteration slamming "
                              "threshold="
                              +str(self.CycleLengthSlamThreshold)+
                              "; choosing a variable to slam")

                        self._pick_one_and_slam_it(ph)
                        self._reset_w_reduce_rho(ph,
                                                 tree_node,
                                                 variable_id,
                                                 full_variable_name)
                    elif (computed_cycle_length > 1) and \
                            (computed_cycle_length \
                                 < self.CycleLengthSlamThreshold):
                        # there was a (potential) cycle, but
                        # the slam threshold wasn't reached.
                        if (self.Verbose or ph._verbose):
                            print(cycle_msg)
                            print("    Taking no action to break cycle - "
                                  "length="+str(computed_cycle_length)+
                                  " does not exceed slam threshold="
                                  +str(self.CycleLengthSlamThreshold))
                    elif (computed_cycle_length \
                              >= self.CycleLengthSlamThreshold) and \
                              ((ph._current_iteration - self._last_slam_iter) \
                                   > self.PH_Iters_Between_Cycle_Slams):
                        # we could have slammed, but we
                        # recently did and are taking a break
                        # to see if things settle down on
                        # their own.
                        if self.Verbose or ph._verbose:
                            print(cycle_msg)
                            print("    Taking no action to break cycle - "
                                  "length="+str(computed_cycle_length)+
                                  " does exceed slam threshold="
                                  +str(self.CycleLengthSlamThreshold)+
                                  ", but another variable was slammed "
                                  "within the past "
                                  +str(self.PH_Iters_Between_Cycle_Slams)+
                                  " iterations")

        #######

        if (ph._current_iteration > self.SlamAfterIter) and \
           ((ph._current_iteration - self._last_slam_iter) \
            > self.PH_Iters_Between_Cycle_Slams):

            if (self.SlamOnlyIfNotConverging and not self.converger_to_use.isImproving(self.PH_Iters_Between_Cycle_Slams)) or \
                    (not self.SlamOnlyIfNotConverging):
                print("Slamming criteria are satisifed - accelerating convergence")
                self._pick_one_and_slam_it(ph)
                self._just_slammed_ = True
            else:
                self._just_slammed_ = False
        else:
            self._just_slammed_ = False

        ### THIS IS EXPERIMENTAL - CODE MAY BELONG SOMEWHERE ELSE ###
        for stage, tree_node, variable_id, variable_values, is_fixed, is_stale in \
              scenario_tree_node_variables_generator_noinstances(
                  ph._scenario_tree,
                  includeDerivedVariables=False,
                  includeLastStage=False,
                  sort=True):

            last_flip_iter = tree_node._w_last_sign_flip_iter[variable_id]
            flip_duration = ph._current_iteration - last_flip_iter

            if (self.CheckWeightOscillationAfterIter > 0) and \
               (ph._current_iteration >= self.CheckWeightOscillationAfterIter):
                if (last_flip_iter == 0) or \
                   (flip_duration >= self.FixIfWeightOscillationCycleLessThan):
                    pass
                else:
                    if self._slam(ph, tree_node, variable_id) is True:
                        tree_node._w_last_sign_flip_iter[variable_id] = 0
                        return

#==================================================
    def post_iteration_k(self, ph):

        pass

#==================================================
    def post_ph_execution(self, ph):

        if self.fix_converged_discrete_variables_at_exit is True:
            print("WW PH extension: Fixing all discrete variables that "
                  "are converged at termination")
            self._fix_all_converged_discrete_variables(ph)

#=========================
    def Int_If_Close_Enough(self, ph, x):
        # if x is close enough to the nearest integer, return the integer
        # else return x
        if abs(round(x)-x) <= ph._integer_tolerance:
            return int(round(x))
        else:
            return x

#=========================
    def _int_convergence_tracking(self,
                                  ph,
                                  tree_node,
                                  variable_id,
                                  node_min,
                                  node_max):

        # keep track of cumulative iters of convergence to the same int
        if (node_min == node_max) and (type(node_min) is int):
            if node_min == tree_node._last_converged_val[variable_id]:
                tree_node._num_iters_converged[variable_id] += 1
            else:
                tree_node._num_iters_converged[variable_id] = 1
                tree_node._last_converged_val[variable_id] = node_min
        else:
            tree_node._num_iters_converged[variable_id] = 0
            tree_node._last_converged_val[variable_id] = 0.5

#=========================
    def _continuous_convergence_tracking(self,
                                         ph,
                                         tree_node,
                                         variable_id,
                                         node_min,
                                         node_max):
        # keep track of cumulative iters of convergence to the same
        # value within tolerance.
        if abs(node_max - node_min) <= ph._integer_tolerance:
            if abs(node_min - tree_node._last_converged_val[variable_id]) \
               <= ph._integer_tolerance:
                tree_node._num_iters_converged[variable_id] = \
                    tree_node._num_iters_converged[variable_id] + 1
            else:
                tree_node._num_iters_converged[variable_id] = 1
                tree_node._last_converged_val[variable_id] = node_min
        else:
            tree_node._num_iters_converged[variable_id] = 0
            # TBD - avoid the magic constant!
            tree_node._last_converged_val[variable_id] = 0.2342343243223423

#=========================
    def _w_history_accounting(self, ph, tree_node, variable_id):
        # do the w hash accounting work
        # we hash on the variable ph weights, and not the values; the
        # latter may not shift for some time, while the former should.
        self.W_hash_rand_val = self.W_hash_seed

        new_sign_vector = []
        old_sign_vector = tree_node._w_sign_vector[variable_id]

        for scenario in tree_node._scenarios:
            weight_value = scenario._w[tree_node._name][variable_id]
            weight_sign = True
            if weight_value < 0.0:
                weight_sign = False
            tree_node._w_hash[variable_id,ph._current_iteration] += \
                weight_value * self.W_hash_rand_val

            new_sign_vector.append(weight_sign)
            self.W_hash_rand_val = \
                (self.W_hash_b + self.W_hash_a * self.W_hash_rand_val) \
                % self.W_hash_c

        hash_val = tree_node._w_hash[variable_id,ph._current_iteration]

        num_flips = 0
        for i in xrange(0,len(old_sign_vector)):
            if new_sign_vector[i] != old_sign_vector[i]:
                num_flips += 1

        tree_node._w_sign_vector[variable_id] = new_sign_vector

        if num_flips >= 1:
            tree_node._w_last_sign_flip_iter[variable_id] = ph._current_iteration

    def compute_cycle_length(self,
                             ph,
                             tree_node,
                             variable_id,
                             report_possible_cycles):

        # return cycles back to closest hash hit for hashval or 0 if
        # no hash hit

        # if the values are converged, then don't report a cycle -
        # often, the weights at convergence are 0s, and even if they
        # aren't, they won't move if the values are uniform.
        if (tree_node._num_iters_converged[variable_id] == 0) and \
           (variable_id not in tree_node._fixed):
            current_hash_value = None
            current_hash_value = \
                tree_node._w_hash[variable_id,ph._current_iteration]
            # scan starting from the farthest point back in history to
            # the closest - this is required to identify the longest
            # possible cycles, which is what we want.
            for i in xrange(max(ph._current_iteration-self.W_hash_history_len-1,1),
                            ph._current_iteration - 1, 1):
                this_hash_value = tree_node._w_hash[variable_id,i]
                if current_hash_value == 0.0:
                    relative_diff = float("inf")
                else:
                    relative_diff = math.fabs(this_hash_value - current_hash_value) / math.fabs(current_hash_value) * 100.0
                # we compare w hash values relatively, as opposed to absolutely.
                if relative_diff <= self.WHashComparisonTolerance:
                    if report_possible_cycles is True:
                        variable_name, index = tree_node._variable_ids[variable_id]
                        print("        Possible cycle detected via PH weight hashing - "
                              "variable="+variable_name+indexToString(index)+
                              " node="+ tree_node._name)
                    msg = ("        Current hash value="+str(current_hash_value)+
                           " matched (within tolerance) hash value="
                           +str(this_hash_value)+" found at PH iteration="
                           +str(i)+"; cycle length="
                           +str(ph._current_iteration - i))

                    if self.ReportCycleWHashDetail:
                        variable_name, index = tree_node._variable_ids[variable_id]
                        print("Cycle W hash history detail for variable="+variable_name+indexToString(index))
                        print("Current hash value: %f" % current_hash_value)
                        print("Iteration  Hash Value   % Diff. Relative to Current Hash Value")
                        for j in xrange(max(ph._current_iteration-self.W_hash_history_len-1,1),
                                        ph._current_iteration - 1, 1):
                            iteration_j_hash_value = tree_node._w_hash[variable_id,j]
                            relative_diff = math.fabs(iteration_j_hash_value - current_hash_value)/math.fabs(current_hash_value)
                            print("%4d %16.6f %10.4f" % (j, iteration_j_hash_value, relative_diff*100.0))
                    return ph._current_iteration - i, msg
        return 0, ""

    def _fix_var(self, ph, tree_node, variable_id, fix_value):

        # fix the variable, account for it and maybe output some trace
        # information
        # note: whether you fix at current values or not can severly
        # impact feasibility later in the game. my original thought
        # was below - but that didn't work out. if you make integers,
        # well, integers, all appears to be well.
        # IMPT: we are leaving the values for individual variables
        #       alone, at potentially smallish and heterogeneous
        #       values. if we fix/round, we run the risk of
        #       infeasibilities due to numerical issues. the computed
        #       value below is strictly for output purposes.
        #       dlw note:
        #       as of aug 1 '09, node_min and node_max should be int
        #       if they should be (so to speak)

        variable_name, index = tree_node._variable_ids[variable_id]
        if self.Verbose or ph._verbose:
            print("        Fixing variable="+variable_name+indexToString(index)+
                  " at tree node="+tree_node._name+" to value="
                  +str(fix_value)+"; converged for "
                  +str(tree_node._num_iters_converged[variable_id])+
                  " iterations")

        tree_node.fix_variable(variable_id, fix_value)

        if tree_node.is_variable_discrete(variable_id):
            self.cumulative_discrete_fixed_count += 1
        else:
            self.cumulative_continuous_fixed_count += 1

        # set the global state flag indicator
        self.variable_fixed = True

    # the last 3 input arguments are the number of iterations the
    # variable is required to be at the respective bound (or lack
    # thereof) before fixing can be initiated.
    def _should_fix_discrete_due_to_conv(self,
                                         tree_node,
                                         variable_id,
                                         lb_iters,
                                         ub_iters,
                                         nb_iters):
        # return True if this should be fixed due to convergence

        # jpw: i don't think this logic is correct - shouldn't
        #   "non-bound" be moved after the lb/ub checks - this doesn't
        #   check a bound!
        # dlw reply: i meant it to mean "without regard to bound" so i
        #   have updated the document
        if nb_iters > 0 and tree_node._num_iters_converged[variable_id] >= nb_iters:
            return True
        else:
            # there is a possibility that the variable doesn't have a
            # bound specified, in which case we should obviously
            # ignore the corresponding lb_iters/ub_iters/nb_iters -
            # which
            # should be none as well!
            (lb, ub) = tree_node._variable_bounds[variable_id]
            conval = tree_node._last_converged_val[variable_id]
            # note: if they are converged node_max == node_min
            if (lb is not None) and (lb_iters > 0) and \
               (tree_node._num_iters_converged[variable_id] >= lb_iters) and \
               (conval == lb):
                return True
            elif (ub is not None) and (ub_iters > 0) and \
                 (tree_node._num_iters_converged[variable_id] >= ub_iters) and \
                 (conval == ub):
                return True
        # if we are still here, nothing triggered fixing
        return False

#=========================
    def _should_fix_continuous_due_to_conv(self, tree_node, variable_id):

        if self.fix_continuous_variables:
            if (self.FixWhenItersConvergedContinuous > 0) and \
               (tree_node._num_iters_converged[variable_id] \
                >= self.FixWhenItersConvergedContinuous):

                return True

        # if we are still here, nothing triggered fixing
        return False

#=========================
    def _slam(self, ph, tree_node, variable_id):
        # this function returns a boolean indicating if it slammed
        # TBD: in the distant future: also: slam it to somewhere it
        #      sort of wants to go
        # e.g., the anywhere case could be to the mode or if more than
        #   one dest is True, pick the one closest to the average as
        #   of sept 09, it is written with the implicit assumption
        #   that only one destination is True or that if not, then
        #   min/max trumps lb/ub and anywhere trumps all

        variable_name, index = tree_node._variable_ids[variable_id]
        lb, ub = tree_node._variable_bounds[variable_id]

        if tree_node.is_variable_discrete(variable_id):
            node_min = \
                self.Int_If_Close_Enough(
                    ph,
                    tree_node._minimums[variable_id])
            node_max = \
                self.Int_If_Close_Enough(
                    ph,
                    tree_node._maximums[variable_id])
            anywhere = round(tree_node._averages[variable_id])
        else:
            node_min = tree_node._minimums[variable_id]
            node_max = tree_node._maximums[variable_id]
            anywhere = tree_node._averages[variable_id]

        slam_basis_string = ""
        fix_value = None
        if self._suffixes['CanSlamToLB'].get(variable_id,
                                             self.CanSlamToLB):
            fix_value = lb
            slam_basis_string = "lower bound"
        if self._suffixes['CanSlamToMin'].get(variable_id,
                                              self.CanSlamToMin):
            fix_value = node_min
            slam_basis_string = "node minimum"
        if self._suffixes['CanSlamToUB'].get(variable_id,
                                             self.CanSlamToUB):
            fix_value = ub
            slam_basis_string = "upper bound"
        if self._suffixes['CanSlamToMax'].get(variable_id,
                                              self.CanSlamToMax):
            fix_value = node_max
            slam_basis_string = "node maximum"
        if self._suffixes['CanSlamToAnywhere'].get(variable_id,
                                                   self.CanSlamToAnywhere):
            fix_value = anywhere
            slam_basis_string = "node average (anywhere)"
        if fix_value is None:
            print("        Warning: Not allowed to slam variable="
                  +variable_name+indexToString(index)+" at tree node="
                  +tree_node._name)
            return False
        else:
            print("        Slamming variable="+variable_name+indexToString(index)+
                  " at tree node="+tree_node._name+" to value="
                  +str(fix_value)+"; value="+slam_basis_string)
            self._fix_var(ph, tree_node, variable_id, fix_value)
            return True

#=========================
    def _pick_one_and_slam_it(self, ph):

        my_stage_suffix = self._variable_stage

        for variable_id in self.slam_list:

            # did we find at least one node to slam in?
            didone = False;

            # it is possible (even likely) that the slam list contains
            # variable values that reside in the final stage - which,
            # because of the initialization loops in the
            # post_ph_initialization() method, will not have a _stage
            # attribute defined.  check for the presence of this
            # attribute and skip if not present, as it doesn't make
            # sense to slam variable values in the final stage anyway.
            if variable_id in my_stage_suffix:
                variable_stage = my_stage_suffix[variable_id]
                for tree_node in variable_stage._tree_nodes:
                    # Find the correct tree node
                    if variable_id in tree_node._variable_ids:
                        # determine if the variable is already fixed (the
                        # trusting version...).
                        if not tree_node.is_variable_fixed(variable_id):
                            didone = self._slam(ph, tree_node, variable_id)
                        break
                if didone:
                    self._last_slam_iter = ph._current_iteration
                    return

        if self.Verbose or ph._verbose:
            print("        Warning: Nothing free with a non-zero slam priority - "
                  "no variable will be slammed")

        # DLW says: Hey, look at this: if we were to start deleting
        # from the slam list this would be wrong."
        if len(self.slam_list) == 0:
            if self.Verbose or ph._verbose:
                print("        (No Slamming Priorities were specified in "
                      "a suffix file.)")

    def _reset_w_reduce_rho(self, ph, tree_node, variable_id, variable_name):

        # only output the rho reduction information if the rho factor
        # will actually have an impact (i.e., it is non-unity).
        if self.Verbose or ph._verbose:
            print("        Resetting W for variable="+variable_name+" to zero"),
            if self.RhoReductionFactor != 1.0:
                print (" and multiplying rho by the RhoReductionFactor="
                  +str(self.RhoReductionFactor)),

        # reset w to zero and reduce rho
        average_rho_value = 0.0
        for scenario in tree_node._scenarios:
            scenario._w[tree_node._name][variable_id] = 0.0
            scenario._rho[tree_node._name][variable_id] *= self.RhoReductionFactor
            average_rho_value += scenario._rho[tree_node._name][variable_id]
        average_rho_value /= float(len(tree_node._scenarios))
        if self.Verbose or ph._verbose:
            if self.RhoReductionFactor != 1.0:
                print("- new average (across all scenarios) rho value="+str(average_rho_value)),

        if self.ScenarioRhoPerturbation != 0.0:
            if self.Verbose or ph._verbose:
                print("; perturbing rho per-scenario")
            lb = 1.0 - self.ScenarioRhoPerturbation
            ub = 1.0 + self.ScenarioRhoPerturbation
            for scenario in tree_node._scenarios:
                scenario._rho[tree_node._name][variable_id] *= random.uniform(lb,ub)

        if self.Verbose or ph._verbose:
            print("")

    # a simple utility to fix any discrete variables to their common
    # value, assuming they are at a common value
    def _fix_all_converged_discrete_variables(self, ph):

        num_variables_fixed = 0

        for stage, tree_node, variable_id, variable_values, is_fixed, is_stale in \
              scenario_tree_node_variables_generator_noinstances(
                  ph._scenario_tree,
                  includeDerivedVariables=False,
                  includeLastStage=False):

            # if the variable is stale, don't waste time fixing and
            # cycle checking. for one, the code will crash :-) due to
            # None values observed during the cycle checking
            # computation.
            if (is_stale is False) and (is_fixed is False):

                if tree_node.is_variable_discrete(variable_id):
                    node_min = \
                        self.Int_If_Close_Enough(ph,
                                                 tree_node._minimums[variable_id])
                    node_max = \
                        self.Int_If_Close_Enough(ph,
                                                 tree_node._maximums[variable_id])

                    if node_min == node_max:
                        self._fix_var(ph, tree_node, variable_id, node_min)
                        num_variables_fixed += 1

        print("WW PH Extension: Total number of variables fixed at PH termination "
              "due to convergence="+str(num_variables_fixed))



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

from pyutilib.misc.config import ConfigBlock, ConfigList, ConfigValue
from pyomo.opt.base.problem import *


@ProblemConfigFactory.register('default')
class Default_Config(BaseProblemConfig):

    def config_block(self, init=False):
        config, blocks = minlp_config_block(init=init)
        return config, blocks


def minlp_config_block(init=False):
    config = ConfigBlock("Configuration for a canonical model construction and optimization sequence")
    blocks={}
       
    #
    # Data
    #
    data = config.declare('data', ConfigBlock())
    data.declare('files', ConfigList(
                [],
                ConfigValue(None, str, 'Filename', None),
                'Model data files',
                None) )
    data.declare('namespaces', ConfigList(
                [],
                ConfigValue(None, str, 'Namespace', None),
                'A namespace that is used to select data in Pyomo data files.',
                None) ).declare_as_argument('--namespace', dest='namespaces', action='append')
    blocks['data'] = data

    #
    # Model
    #
    model = config.declare('model', ConfigBlock())
    model.declare('filename', ConfigValue(
                None,
                str,
                'The Python module that specifies the model',
                None ) )
    model.declare('object name', ConfigValue(
                None, 
                str,
                'The name of the model object that is created in the specified Pyomo module',
                None ) ).declare_as_argument('--model-name', dest='model_name') 
    model.declare('type', ConfigValue(
                None,
                str,
                'The problem type',
                None ) )
    model.declare('options', ConfigBlock(
                implicit=True,
                description='Options used to construct the model') )
    model.declare('linearize expressions', ConfigValue(
                False,
                bool,
                'An option intended for use on linear or mixed-integer models in which expression trees in a model (constraints or objectives) are compacted into a more memory-efficient and concise form.',
                None) )
    model.declare('save file', ConfigValue(
                None,
                str,
                "The filename to which the model is saved. The suffix of this filename specifies the file format.",
                None) )
    model.declare('save format', ConfigValue(
                None,
                str,
                "The format that the model is saved. When specified, this overrides the format implied by the 'save file' option.",
                None) )
    model.declare('symbolic solver labels', ConfigValue(
                False,
                bool,
                'When interfacing with the solver, use symbol names derived from the model. For example, \"my_special_variable[1_2_3]\" instead of \"v1\". Useful for debugging. When using the ASL interface (--solver-io=nl), generates corresponding .row (constraints) and .col (variables) files. The ordering in these files provides a mapping from ASL index to symbolic model names.',
                None) ).declare_as_argument(dest='symbolic_solver_labels')
    model.declare('file determinism', ConfigValue(
                1,
                int,
                'When interfacing with a solver using file based I/O, set the effort level for ensuring the file creation process is determistic. The default (1) sorts the index of components when transforming the model. Anything less than 1 disables index sorting. Anything greater than 1 additionaly sorts by component name to override declartion order.',
                None) ).declare_as_argument(dest='file_determinism')
    blocks['model'] = model

    #
    # Transform
    #
    transform = ConfigBlock()
    transform.declare('name', ConfigValue(
                None, 
                str,
                'Name of the model transformation',
                None ) )
    transform.declare('options', ConfigBlock(
                implicit=True,
                description='Transformation options'
                ) )
    blocks['transform'] = transform
    #
    transform_list = config.declare('transform', ConfigList(
                [],
                ConfigValue(None, str, 'Transformation', None),
                'List of model transformations',
                None) ).declare_as_argument( dest='transformations',
                                             action='append' )
    if init:
        transform_list.append()

    #
    # Preprocess
    #
    config.declare('preprocess', ConfigList(
                [],
                ConfigValue(None, str, 'Module', None),
                'Specify a Python module that gets immediately executed (before the optimization model is setup).',
                None) ).declare_as_argument(dest='preprocess')

    #
    # Runtime
    #
    runtime = config.declare('runtime', ConfigBlock())
    runtime.declare('logging', ConfigValue(
                None, 
                str,
                'Logging level:  quiet, warning, info, verbose, debug',
                None) ).declare_as_argument(dest="logging", metavar="LEVEL")
    runtime.declare('logfile', ConfigValue(
                None, 
                str,
                'Redirect output to the specified file.',
                None) ).declare_as_argument(dest="output", metavar="FILE")
    runtime.declare('catch errors', ConfigValue(
                False, 
                bool,
                'Trigger failures for exceptions to print the program stack.',
                None) ).declare_as_argument('-c', '--catch-errors', dest="catch")
    runtime.declare('disable gc', ConfigValue(
                False, 
                bool,
                'Disable the garbage collecter.',
                None) ).declare_as_argument('--disable-gc', dest='disable_gc')
    runtime.declare('interactive', ConfigValue(
                False, 
                bool,
                'After executing Pyomo, launch an interactive Python shell. If IPython is installed, this shell is an IPython shell.',
                None) )
    runtime.declare('keep files', ConfigValue(
                False, 
                bool,
                'Keep temporary files',
                None) ).declare_as_argument('-k', '--keepfiles', dest='keepfiles')
    runtime.declare('paths', ConfigList(
                [], 
                ConfigValue(None, str, 'Path', None),
                'Give a path that is used to find the Pyomo python files.',
                None) ).declare_as_argument('--path', dest='path')
    runtime.declare('profile count', ConfigValue(
                0, 
                int,
                'Enable profiling of Python code. The value of this option is the number of functions that are summarized.',
                None) ).declare_as_argument(dest='profile_count', metavar='COUNT')
    runtime.declare('profile memory', ConfigValue(
                0, 
                int,
                "Report memory usage statistics for the generated instance and any associated processing steps. A value of 0 indicates disabled. A value of 1 forces the print of the total memory after major stages of the pyomo script. A value of 2 forces summary memory statistics after major stages of the pyomo script. A value of 3 forces detailed memory statistics during instance creation and various steps of preprocessing. Values equal to 4 and higher currently provide no additional information. Higher values automatically enable all functionality associated with lower values, e.g., 3 turns on detailed and summary statistics.",
                None) )
    runtime.declare('report timing', ConfigValue(
                False, 
                bool,
                'Report various timing statistics during model construction.',
                None) ).declare_as_argument(dest='report_timing')
    runtime.declare('tempdir', ConfigValue(
                None, 
                str,
                'Specify the directory where temporary files are generated.',
                None) ).declare_as_argument(dest='tempdir')
    blocks['runtime'] = runtime
    #
    return config, blocks




from pyomo.core import *
from pyomo.pysp.annotations import (StochasticConstraintBoundsAnnotation,
                                    StochasticConstraintBodyAnnotation,
                                    StochasticObjectiveAnnotation)

d = {}
d['Scenario1'] = 0
d['Scenario2'] = 1
d['Scenario3'] = 2
def create_instance(scenario_name):
    cnt = d[scenario_name]

    model = ConcreteModel()
    # first stage
    model.x = Var(bounds=(0,10))
    # first stage derived
    model.y = Expression(expr=model.x + 1)
    model.fx = Var()
    # second stage
    model.z = Var(bounds=(-10, 10))
    # second stage derived
    model.q = Expression(expr=model.z**2)
    model.fz = Var()
    model.r = Var()
    # stage costs
    model.StageCost = Expression([1,2])
    model.StageCost.add(1, model.fx)
    model.StageCost.add(2, -model.fz + model.r - cnt)
    model.o = Objective(expr=sum_product(model.StageCost))

    model.ZERO = Param(initialize=0, mutable=True)
    if cnt == 0:
        cnt = model.ZERO

    model.c_first_stage = Constraint(expr= model.x >= 0)

    # test our handling of intermediate variables that
    # are created by Piecewise but can not necessarily
    # be declared on the scenario tree
    model.p_first_stage = Piecewise(model.fx, model.x,
                                    pw_pts=[0.,2.,5.,7.,10.],
                                    pw_constr_type='EQ',
                                    pw_repn='INC',
                                    f_rule=[10.,10.,9.,10.,10.],
                                    force_pw=True)

    model.c_second_stage = Constraint(expr= model.x + model.r * cnt >= -100)
    model.r_second_stage = Constraint(expr= inequality(-cnt, model.r, 0))
    # exercise more of the code by making this an indexed
    # block
    model.p_second_stage = Piecewise([1], model.fz, model.z,
                                     pw_pts=[-10,-5.,0.,5.,10.],
                                     pw_constr_type='EQ',
                                     pw_repn='INC',
                                     f_rule=[0.,0.,-1.,2.+cnt,1.],
                                     force_pw=True)

    return model

def pysp_instance_creation_callback(scenario_name, node_names):

    model = create_instance(scenario_name)

    #
    # SMPS Related Annotations
    #
    model.stoch_rhs = StochasticConstraintBoundsAnnotation()
    # declarations can be blocks to imply that all
    # components (Constraints in this case) should be
    # considered on that block
    model.stoch_rhs.declare(model.p_second_stage)
    model.stoch_rhs.declare(model.r_second_stage)
    model.stoch_matrix = StochasticConstraintBodyAnnotation()
    # exercise more of the code by testing this with an
    # indexed block and a single block
    model.stoch_matrix.declare(model.c_second_stage, variables=[model.r])
    model.stoch_matrix.declare(model.p_second_stage[1])
    model.stoch_objective = StochasticObjectiveAnnotation()

    return model



#! /usr/bin/env python
#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

#
# This is virgin parametric lagrange multiplier search
# no pre-processing and no gap-closing...just testing algorithm
#

import sys
import time
import datetime
import operator
# used for debug when Result was not returning correctly
import inspect

from pyomo.pysp.scenariotree.instance_factory import \
   ScenarioTreeInstanceFactory
from pyomo.pysp.phinit import (construct_ph_options_parser,
                               GenerateScenarioTreeForPH,
                               PHFromScratch)
from pyomo.pysp.phutils import find_active_objective
from pyomo.pysp import lagrangeutils as lagrUtil

##############################################################

def datetime_string():
   return "datetime = "+str(datetime.datetime.now())

###################################
def run(args=None):
###################################

   print("RUNNING - run args=%s" % str(args))

   import pyomo.environ

   def LagrangeParametric(args=None):
      class Object(object): pass
      Result = Object()
      Result.status = 'LagrangeParam begins '+ datetime_string() + '...running new ph'
      ph = None

      blanks = "                          "  # used for formatting print statements
# options used
      betaMin       = options.beta_min
      betaMax       = options.beta_max
      betaTol       = options.beta_tol
      gapTol        = options.Lagrange_gap
      minProb       = options.min_prob
      maxIntervals  = options.max_intervals
      maxTime       = options.max_time
      IndVarName    = options.indicator_var_name
      multName      = options.lambda_parm_name
      CCStageNum    = options.stage_num
      csvPrefix     = options.csvPrefix
      verbosity     = options.verbosity
      verbosity = 2 # override for debug (= 3 to get super-debug)
      HGdebug = 0   # special debug (not public)
# local...may become option
      optTol = gapTol
####################################################################
      STARTTIME = time.time()

      Result.status = "options set"
      if verbosity > 1:
        print("From LagrangeParametric, status = %s\tSTARTTIME = %s" \
                % (str(getattr(Result,'status')), str(STARTTIME)))

      ph = PHFromScratch(options)
      Result.ph = ph
      rootnode = ph._scenario_tree._stages[0]._tree_nodes[0]   # use rootnode to loop over scenarios
      ReferenceInstance = ph._instances[rootnode._scenarios[0]._name]  # arbitrary scenario

      if find_active_objective(ph._scenario_tree._scenarios[0]._instance,safety_checks=True).is_minimizing():
         sense = 'min'
      else:
         sense = 'max'

      scenario_count = len(full_scenario_tree._stages[-1]._tree_nodes)
      if options.verbosity > 0: print("%s %s scenarios" % (str(sense),str(scenario_count)))

# initialize
      Result.status = 'starting at '+datetime_string()
      if verbosity > 0:
         print(Result.status)
      ScenarioList = []
      lambdaval = 0.
      lagrUtil.Set_ParmValue(ph, multName,lambdaval)

      # IMPORTANT: Preprocess the scenario instances
      #            before fixing variables, otherwise they
      #            will be preprocessed out of the expressions
      #            and the output_fixed_variable_bounds option
      #            will have no effect when we update the
      #            fixed variable values (and then assume we
      #            do not need to preprocess again because
      #            of this option).
      ph._preprocess_scenario_instances()

      sumprob = 0.
      minprob = 1.
      maxprob = 0.
      # fixed = 0 to get PR point at b=0
      lagrUtil.FixAllIndicatorVariables(ph, IndVarName, 0)
      for scenario in rootnode._scenarios:
         instance = ph._instances[scenario._name]
         sname = scenario._name
         sprob = scenario._probability
         sumprob = sumprob + sprob
         minprob = min(minprob,sprob)
         maxprob = max(maxprob,sprob)
         ScenarioList.append([sname,sprob])

      ScenarioList.sort(key=operator.itemgetter(1))   # sorts from min to max probability
      if verbosity > 0:
         print("probabilities sum to %f range: %f to %f" % (sumprob,minprob,maxprob))
      Result.ScenarioList = ScenarioList

# Write ScenarioList = name, probability in csv file sorted by probability
      outName = csvPrefix + 'ScenarioList.csv'
      print("writing to %s" % outName)
      with open(outName,'w') as outFile:
         for scenario in ScenarioList:
            outFile.write(scenario[0]+", "+str(scenario[1])+'\n')
      Result.ScenarioList = ScenarioList

      addstatus = 'Scenario List written to ' + csvPrefix+'ScenarioList.csv'
      Result.status = Result.status + '\n' + addstatus
      if verbosity > 0:
         print(addstatus)

      if verbosity > 0:
         print("solve begins %s" % datetime_string())
         print("\t- lambda = %f" % lambdaval)
      SolStat, zL = lagrUtil.solve_ph_code(ph, options)
      if verbosity > 0:
         print("solve ends %s" % datetime_string())
         print("\t- status = %s" % str(SolStat))
         print("\t- zL = %s" % str(zL))

      bL = Compute_ExpectationforVariable(ph, IndVarName, CCStageNum)
      if bL > 0:
         print("** bL = %s > 0 (all %s = 0)" % (str(bL), str(IndVarName)))
         return Result

      if verbosity > 0:  print("Initial optimal obj = %s for bL = %s" % (str(zL), str(bL)))

      # fixed = 1 to get PR point at b=1
      lagrUtil.FixAllIndicatorVariables(ph, IndVarName, 1)

      if verbosity > 0:
        print("solve begins %s" % datetime_string())
        print("\t- lambda = %s" % str(lambdaval))
      SolStat, zU = lagrUtil.solve_ph_code(ph, options)
      if verbosity > 0:
        print("solve ends %s" % datetime_string())
        print("\t- status = %s" % str(SolStat))
        print("\t- zU = %s" % str(zU))
      if not SolStat[0:2] == 'ok':
         print(str(SolStat[0:3])+" is not 'ok'")
         addstatus = "** Solution is non-optimal...aborting"
         print(addstatus)
         Result.status = Result.status + "\n" + addstatus
         return Result

      bU = Compute_ExpectationforVariable(ph, IndVarName, CCStageNum)
      if bU < 1.- betaTol and verbosity > 0:
         print("** Warning:  bU = %s  < 1" % str(bU))

### enumerate points in PR space (all but one scenario)
#      Result.lbz = [ [0,bL,zL], [None,bU,zU] ]
#      for scenario in rootnode._scenarios:
#         sname = scenario._name
#         instance = ph._instances[sname]
#         print "excluding scenario",sname
#         getattr(instance,IndVarName).value = 0
#         print sname,"value =",getattr(instance,IndVarName).value,getattr(instance,IndVarName).fixed
#         SolStat, z = lagrUtil.solve_ph_code(ph, options)
#         b = Compute_ExpectationforVariable(ph, IndVarName, CCStageNum)
#         print "solve ends with status =",SolStat,"(b, z) =",b,z
#         getattr(instance,IndVarName).value = 1
#         Result.lbz.append([None,b,z])
#         for t in instance.TimePeriods:
#           print "Global at",t,"=",instance.posGlobalLoadGenerateMismatch[t].value, \
#                '-',instance.negGlobalLoadGenerateMismatch[t].value,"=",\
#                    instance.GlobalLoadGenerateMismatch[t].value,\
#               "\tDemand =",instance.TotalDemand[t].value, ",",\
#                "Reserve =",instance.ReserveRequirement[t].value
#
#      PrintPRpoints(Result.lbz)
#      return Result
#### end enumeration
########################################################################

      if verbosity > 1:
         print("We have bU = %s ...about to free all %s for %d scenarios" % \
                (str(bU), str(IndVarName), len(ScenarioList)))

      # free scenario selection variable
      lagrUtil.FreeAllIndicatorVariables(ph, IndVarName)

      if verbosity > 1:
         print("\tall %s freed; elapsed time = %f" % (str(IndVarName), time.time() - STARTTIME))

# initialize with the two endpoints
      Result.lbz = [ [0.,bL,zL], [None,bU,zU] ]
      Result.selections = [[], ScenarioList]
      NumIntervals = 1
      if verbosity > 0:
         print("Initial relative Lagrangian gap = %f maxIntervals = %d" % (1-zL/zU, maxIntervals))
         if verbosity > 1:
            print("entering while loop %s" % datetime_string())
         print("\n")

############ main loop to search intervals #############
########################################################
      while NumIntervals < maxIntervals:
         lapsedTime = time.time() - STARTTIME
         if lapsedTime > maxTime:
            addstatus = '** max time reached ' + str(lapsedTime)
            print(addstatus)
            Result.status = Result.status + '\n' + addstatus
            break
         if verbosity > 1:
            print("Top of while with %d intervals elapsed time = %f" % (NumIntervals, lapsedTime))
            PrintPRpoints(Result.lbz)

         lambdaval = None
### loop over PR points to find first unfathomed interval to search ###
         for PRpoint in range(1,len(Result.lbz)):
            if Result.lbz[PRpoint][0] == None:
# multiplier = None means interval with upper endpoint at PRpoint not fathomed
               bL = Result.lbz[PRpoint-1][1]
               zL = Result.lbz[PRpoint-1][2]
               bU = Result.lbz[PRpoint][1]
               zU = Result.lbz[PRpoint][2]
               lambdaval = (zU - zL) / (bU - bL)
               break

#############################
# Exited from the for loop
         if verbosity > 1:
            print("exited for loop with PRpoint = %s ...lambdaval = %s" % (PRpoint, lambdaval))
         if lambdaval == None: break # all intervals are fathomed

         if verbosity > 1: PrintPRpoints(Result.lbz)
         if verbosity > 0:
            print("Searching for b in [%s, %s] with %s = %f" % (str(round(bL,4)), str(round(bU,4)), multName, lambdaval))

# search interval (bL,bU)
         lagrUtil.Set_ParmValue(ph, multName,lambdaval)
         if verbosity > 0:
            print("solve begins %s" % datetime_string())
            print("\t- %s = %f" % (multName, lambdaval))

         #########################################################
         SolStat, Lagrangian = lagrUtil.solve_ph_code(ph, options)
         #########################################################
         if not SolStat[0:2] == 'ok':
            addstatus = "** Solution status " + SolStat + " is not optimal"
            print(addstatus)
            Result.status = Result.status + "\n" + addstatus
            return Result

         b = Compute_ExpectationforVariable(ph, IndVarName, CCStageNum)
         z = Lagrangian + lambdaval*b
         if verbosity > 0:
            print("solve ends %s" % datetime_string())
            print("\t- Lagrangian = %f" % Lagrangian)
            print("\t- b = %s" % str(b))
            print("\t- z = %s" % str(z))
            print("\n")

# We have PR point (b,z), which may be new or one of the endpoints
##################################################################

######### Begin tolerance tests ##########
# Test that b is in [bL,bU]
         if verbosity > 1: print("\ttesting b")
         if b < bL - betaTol or b > bU + betaTol:
            addstatus = "** fatal error: probability (= " + str(b) + \
                ") is outside interval, (" + str(bL) + ", " + str(bU) + ")"
            addstatus = addstatus + "\n\t(tolerance = " + str(betaTol) + ")"
            print(addstatus+'\n')
            Result.status = Result.status + addstatus
            return Result
# Test that z is in [zL,zU]
         if verbosity > 1: print("\ttesting z")
# using optTol as absolute tolerance (not relative)
#   ...if we reconsider, need to allow negative z-values
         if z < zL - optTol or z > zU + optTol:
            addstatus = "** fatal error: obj (= " + str(z) + \
                ") is outside interval, (" + str(zL) + ", " + str(zU) + ")"
            print(addstatus+'\n')
            Result.status = Result.status + addstatus
            return Result

# Ok, we have (b,z) in [(bL,zL), (bU,zU)], at least within tolerances

         oldLagrangian = zL - lambdaval*bL
# ensure lambdaval set such that endpoints have same Lagrangian value
# (this is probably unnecessary, but check anyway)
         if abs(oldLagrangian - (zU - lambdaval*bU)) > optTol*abs(oldLagrangian):
            addstatus = "** fatal error: Lagrangian at (bL,zL) = " + \
                str(oldLagrangian) + " not= " + str(zU-lambdaval*bU) + \
                "\n\t(optTol = " + str(optTol) + ")"
            Result.status = Result.status + addstatus
            return Result

# no more fatal error tests...need to know if (b,z) is endpoint or new

         if verbosity > 1: print("No anomalies...testing if b = bL or bU")

# Test if endpoint is an alternative optimum of Lagrangian
# ...using optTol as *relative* tolerance
# (could use other reference values -- eg, avg or max of old and new Lagrangian values)
         refValue =  max( min( abs(oldLagrangian), abs(Lagrangian) ), 1.)
         alternativeOpt = abs( oldLagrangian - Lagrangian ) <= optTol*refValue

# alternativeOpt = True means we computed point (b,z) is alternative optimum such that:
#   case 1: (b,z) = endpoint, in which case we simply fathom [bL,bU] by setting PRpoint
#            to [lambdaval,bU,zU] (the numeric value of multiplier means fathomed)
#   case 2: (b,z) is new PR point on line segment, in which case we split into
#           [bL,b] and [b,bU], with both fathomed

         if verbosity > 1:
            print("oldLagrangian = %s" % str(oldLagrangian))
            if alternativeOpt: print(":= Lagrangian = %s" % str(Lagrangian ))
            else: print("> Lagrangian = %s" % str(Lagrangian))

         if alternativeOpt:
# setting multiplier of (bU,zU) to a numeric fathoms the interval [bL,bU]
            Result.lbz[PRpoint][0] = lambdaval

# test if (b,z) is an endpoint
         newPRpoint = abs(b-bL) > betaTol and abs(b-bU) > betaTol
         if not newPRpoint:
# ...(b,z) is NOT an endpoint (or sufficiently close), so split and fathom
            if verbosity > 1:
               print("\tnot an endpoint\tlbz = %s" % str(Result.lbz[PRpoint]))
            if verbosity > 0:
               print("Lagangian solution is new PR point on line segment of (" \
                  + str(bL) + ", " + str(bU) +")")
               print("\tsplitting (bL,bU) into (bL,b) and (b,bU), both fathomed")
# note:  else ==> b = bL or bU, so we do nothing, having already fathomed [bL,bU]

# (b,z) is new PR point, so split interval (still in while loop)
##########################################
# alternative optimum ==> split & fathom: (bL,b), (b,bU)
         if verbosity > 1:
            print("\talternativeOpt %s newPRpoint = %s" % (alternativeOpt, newPRpoint))
         if newPRpoint:
            NumIntervals += 1
            if alternativeOpt:
               if verbosity > 1: print("\tInsert [lambdaval,b,z] at %f" % PRpoint)
               Result.lbz = Insert([lambdaval,b,z],PRpoint,Result.lbz)
               addstatus = "Added PR point on line segment of envelope"
               if verbosity > 0: print(addstatus+'\n')
            else:
               if verbosity > 1: print("\tInsert [None,b,z] at %f" % PRpoint)
               Result.lbz = Insert([None,b,z],PRpoint,Result.lbz)
               addstatus = "new envelope extreme point added (interval split, not fathomed)"
            Result.status = Result.status + "\n" + addstatus

            if verbosity > 1:
               print("...after insertion:")
               PrintPRpoints(Result.lbz)

# get the selections of new point (ie, scenarios for which delta=1)
            Selections = []
            for scenario in ScenarioList:
               instance = ph._instances[scenario[0]]
               if getattr(instance,IndVarName).value == 1:
                  Selections.append(scenario)
            Result.selections = Insert(Selections,PRpoint,Result.selections)

            if verbosity > 0:
               print("Interval "+str(PRpoint)+", ["+str(bL)+", "+str(bU)+ \
                 "] split at ("+str(b)+", "+str(z)+")")
               print("\tnew PR point has "+str(len(Selections))+" selections")

            if verbosity > 1: print("test that selections list aligned with lbz")
            if not len(Result.lbz) == len(Result.selections):
               print("** fatal error: lbz not= selections")
               PrintPRpoints(Result.lbz)
               print("Result.selections:")
               for i in range(Result.selections): print("%d %f" % (i,Result.selections[i]))
               return Result

# ok, we have split and/or fathomed interval
         if NumIntervals >= maxIntervals:
# we are about to leave while loop due to...
            addstatus = "** terminating because number of intervals = " + \
                    str(NumIntervals) + " >= max = " + str(maxIntervals)
            if verbosity > 0: print(addstatus+'\n')
            Result.status = Result.status + "\n" + addstatus

# while loop continues
         if verbosity > 1:
            print("bottom of while loop")
            PrintPRpoints(Result.lbz)

###################################################
# end while NumIntervals < maxIntervals:
#     ^ this is indentation of while loop
################ end while loop ###################

      if verbosity > 1:  print("\nend while loop...setting multipliers")
      for i in range(1,len(Result.lbz)):
         db = Result.lbz[i][1] - Result.lbz[i-1][1]
         dz = Result.lbz[i][2] - Result.lbz[i-1][2]
         if dz > 0:
            Result.lbz[i][0] = dz/db
         else:
            #print "dz =",dz," at ",i,": ",Result.lbz[i]," -",Result.lbz[i-1]
            Result.lbz[i][0] = 0
      if verbosity > 0: PrintPRpoints(Result.lbz)

      addstatus = '\nLagrange multiplier search ends'+datetime_string()
      if verbosity > 0:
         print(addstatus+'\n')
      Result.status = Result.status + addstatus

      outName = csvPrefix + "PRoptimal.csv"
      with open(outName,'w') as outFile:
         if verbosity > 0:
            print("writing PR points to "+outName+'\n')
         for lbz in Result.lbz:
            outFile.write(str(lbz[1])+ ", " +str(lbz[2])+'\n')

      outName = csvPrefix + "OptimalSelections.csv"
      with open(outName,'w') as outFile:
         if verbosity > 0:
            print("writing optimal selections for each PR point to "+csvPrefix+'PRoptimal.csv\n')
         for selections in Result.selections:
            char = ""
            thisSelection = ""
            for slist in selections:
               if slist:
                  thisSelection = thisSelection + char + slist[0]
                  char = ","
            outFile.write(thisSelection+'\n')

      if verbosity > 0:
         print("\nReturning status:\n %s \n=======================" % Result.status)

################################
      if verbosity > 2:
         print("\nAbout to return...Result attributes: %d" % len(inspect.getmembers(Result)))
         for attr in inspect.getmembers(Result): print(attr[0])
         print("\n===========================================")
# LagrangeParametric ends here
      return Result
################################


####################################### start run ####################################

   AllInOne = False

########################
# options defined here
########################
   try:
      conf_options_parser = construct_ph_options_parser("lagrange [options]")
      conf_options_parser.add_argument("--beta-min",
                                     help="The min beta level for the chance constraint. Default is 0",
                                     action="store",
                                     dest="beta_min",
                                     type=float,
                                     default=0.)
      conf_options_parser.add_argument("--beta-max",
                                     help="The beta level for the chance constraint. Default is 1.",
                                     action="store",
                                     dest="beta_max",
                                     type=float,
                                     default=1.)
      conf_options_parser.add_argument("--beta-tol",
                                     help="Tolerance for testing equality to beta. Default is 1e-5",
                                     action="store",
                                     dest="beta_tol",
                                     type=float,
                                     default=1e-5)
      conf_options_parser.add_argument("--Lagrange-gap",
                                     help="The (relative) Lagrangian gap acceptable for the chance constraint. Default is 10^-4",
                                     action="store",
                                     type=float,
                                     dest="Lagrange_gap",
                                     default=0.0001)
      conf_options_parser.add_argument("--min-prob",
                                     help="Tolerance for testing probability > 0. Default is 1e-9",
                                     action="store",
                                     dest="min_prob",
                                     type=float,
                                     default=1e-5)
      conf_options_parser.add_argument("--max-intervals",
                                     help="The max number of intervals generated; if causes termination, non-fathomed intervals have multiplier=None.  Default = 100.",
                                     action="store",
                                     dest="max_intervals",
                                     type=int,
                                     default=100)
      conf_options_parser.add_argument("--max-time",
                                     help="Maximum time (seconds). Default is 3600.",
                                     action="store",
                                     dest="max_time",
                                     type=float,
                                     default=3600)
      conf_options_parser.add_argument("--lambda-parm-name",
                                     help="The name of the lambda parameter in the model. Default is lambdaMult",
                                     action="store",
                                     dest="lambda_parm_name",
                                     type=str,
                                     default="lambdaMult")
      conf_options_parser.add_argument("--indicator-var-name",
                                     help="The name of the indicator variable for the chance constraint. The default is delta",
                                     action="store",
                                     dest="indicator_var_name",
                                     type=str,
                                     default="delta")
      conf_options_parser.add_argument("--stage-num",
                                     help="The stage number of the CC indicator variable (number, not name). Default is 2",
                                     action="store",
                                     dest="stage_num",
                                     type=int,
                                     default=2)
      conf_options_parser.add_argument("--csvPrefix",
                                     help="Output file name.  Default is ''",
                                     action="store",
                                     dest="csvPrefix",
                                     type=str,
                                     default='')
      conf_options_parser.add_argument("--verbosity",
                                     help="verbosity=0 is no extra output, =1 is medium, =2 is debug, =3 super-debug. Default is 1.",
                                     action="store",
                                     dest="verbosity",
                                     type=int,
                                     default=1)
# The following needed for solve_ph_code in lagrangeutils
      conf_options_parser.add_argument("--solve-with-ph",
                                     help="Perform solves via PH rather than an EF solve. Default is False",
                                     action="store_true",
                                     dest="solve_with_ph",
                                     default=False)
##HG: deleted params filed as deletedParam.py
#######################################################################################################

      options = conf_options_parser.parse_args(args=args)
      # temporary hack
      options._ef_options = conf_options_parser._ef_options
      options._ef_options.import_argparse(options)
   except SystemExit as _exc:
      # the parser throws a system exit if "-h" is specified - catch
      # it to exit gracefully.
      return _exc.code

   # create the reference instances and the scenario tree - no
   # scenario instances yet.
   if options.verbosity > 0:
        print("Loading reference model and scenario tree")
# Dec 18
#   scenario_instance_factory, full_scenario_tree = load_models(options)
   scenario_instance_factory = \
        ScenarioTreeInstanceFactory(options.model_directory,
                                    options.instance_directory)

   full_scenario_tree = \
            GenerateScenarioTreeForPH(options,
                                      scenario_instance_factory)

####
   try:
      if (scenario_instance_factory is None) or (full_scenario_tree is None):
         raise RuntimeError("***ERROR: Failed to initialize the model and/or scenario tree data.")

      # load_model gets called again, so lets make sure unarchived directories are used
      options.model_directory = scenario_instance_factory._model_filename
      options.instance_directory = scenario_instance_factory._scenario_tree_filename

########## Here is where multiplier search is called from run() ############
      Result = LagrangeParametric()
#####################################################################################
   finally:

      # delete temporary unarchived directories
      scenario_instance_factory.close()

   if options.verbosity > 0:
      print("\n===========================================")
      print("\nreturned from LagrangeParametric")
      if options.verbosity > 2:
         print("\nFrom run, Result should have status and ph objects...")
         for attr in inspect.getmembers(Result): print(attr)
         print("\n===========================================")

   try:
     status = Result.status
     print("status = "+str(Result.status))
   except:
     print("status not defined")
     sys.exit()

   try:
      lbz = Result.lbz
      PrintPRpoints(lbz)
      with open(options.csvPrefix+"PRoptimal.csv",'w') as outFile:
         for lbz in Result.lbz:
            outFile.write(str(lbz[1])+ ", " +str(lbz[2])+'\n')
   except:
      print("Result.lbz not defined")
      sys.exit()

   try:
      ScenarioList = Result.ScenarioList
      ScenarioList.sort(key=operator.itemgetter(1))
      with open(options.csvPrefix+"ScenarioList.csv",'w') as outFile:
         for scenario in ScenarioList:
            outFile.write(scenario[0]+", "+str(scenario[1])+'\n')
   except:
      print("Result.ScenarioList not defined")
      sys.exit()

########################### run ends here ##############################

########## begin function definitions ################
def Compute_ExpectationforVariable(ph, IndVarName, CCStageNum):
   SumSoFar = 0.0
   node_probability = 0.0
   stage = ph._scenario_tree._stages[CCStageNum-1]
   for tree_node in stage._tree_nodes:
      for scenario in tree_node._scenarios:
         instance = ph._instances[scenario._name]
         #print "scenario._probability:",scenario._probability
         node_probability += scenario._probability
         #print "node_probability:",node_probability
         #print "getattr(instance, IndVarName).value:",getattr(instance, IndVarName).value
         SumSoFar += scenario._probability * getattr(instance, IndVarName).value
         #print "SumSoFar:",SumSoFar
   return SumSoFar / node_probability

#######################################

def Insert(newpoint,location,List):
    newList = []
    for i in range(location): newList.append(List[i])
    newList.append(newpoint)
    for i in range(location,len(List)): newList.append(List[i])
    return newList

#######################################

def ismember(List,member):  # designed to test 1st member of each list in List (ie, 1st column)
   for i in List:
      if len(i[0]) == 0: continue   # in case list contains empty list
      if i[0] == member: return True
   return

#######################################

def putcommas(num):
   snum = str(num)
   decimal = snum.find('.')
   if decimal >= 0:
      frac = snum[decimal:]
      snum = snum[0:decimal]
   else: frac = ''
   if len(snum) < 4: return snum + frac
   else: return putcommas(snum[:len(snum)-3]) + "," + snum[len(snum)-3:len(snum)] + frac

#######################################

def HGdebugSolve(ph,options,ScenarioList):
   print("HGdebugSolve prints attributes after solving")
   SolStat, Lagrangian = lagrUtil.solve_ph_code(ph, options)
   IndVarName = options.indicator_var_name
   for scenario in ScenarioList:
       sname = scenario[0]
       instance = ph._instances[sname]
       if getattr(instance,IndVarName).fixed: stat = "fixed"
       else: stat = "free"
       lambdaval = getattr(instance, options.lambda_parm_name).value
       #print IndVarName+"("+sname+") =",getattr(instance,IndVarName).value, stat, "\tlambda =",lambdaval
   b = Compute_ExpectationforVariable(ph, IndVarName, options.stage_num)
   z = Lagrangian + lambdaval*b
   #print "L = ",z," - ",lambdaval,"x",b
   return SolStat, Lagrangian

def PrintPRpoints(PRlist):
   if len(PRlist) == 0:
      print("No PR points")
   else:
      print("%d PR points:" % len(PRlist))
      blanks = "                      "
      print("            lambda        beta-probability       min cost")
      for row in PRlist:
         b = float(round(row[1],4))
         z = float(round(row[2]))
# lambda = row[0] could be float, string, or None
         sl = str(row[0])
         sl = blanks[0:20-len(sl)] + sl
         sb = str(b)
         sb = blanks[0:20-len(sb)] + sb
         sz = putcommas(z)
         sz = blanks[2:20-len(sz)] + sz
         print(sl+" "+sb+" "+sz)
      print("==================================================================")
   return

#
# the main script routine starts here
#

if __name__ == "__main__":

   run()

# RESTORE THE BELOW ASAP
#try:
#    run()
#except ValueError, str:
#    print "VALUE ERROR:"
#    print str
#except IOError, str:
#    print "IO ERROR:"
#    print str
#except pyutilib.common.ApplicationError, str:
#    print "APPLICATION ERROR:"
#    print str
#except RuntimeError, str:
#    print "RUN-TIME ERROR:"
##    print str
#except:
#   print "Encountered unhandled exception", sys.exc_info()[0]
#   traceback.print_exc()



#
# a model that is missing a declaration of
# a stochastic constraint variable
#
from pyomo.pysp.tests.convert.utils import *

pysp_scenario_tree_model_callback = \
    simple_twostage_scenario_tree

def pysp_instance_creation_callback(scenario_name, node_names):
    model = simple_twostage_model()
    if scenario_name == "Scenario1":
        model.cc = Constraint(expr=model.x + 2*model.y >= 1)
    elif scenario_name == "Scenario2":
        model.cc = Constraint(expr=model.x + 3*model.y >= 1)
    else:
        assert False
    model.smat = StochasticConstraintBodyAnnotation()
    model.smat.declare(model.cc,
                       variables=(model.x,))
    return model



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

import code
import sys
import os
import string
import signal
import subprocess
import pyutilib.subprocess

import pyutilib.pyro
from pyutilib.misc import Options
from pyomo.opt import SolverResults
from pyomo.common._command import pyomo_command
import pyomo.scripting.pyomo_parser

@pyomo_command('pyomo_ns', "Launch a Pyro name server for Pyomo")
def pyomo_ns():
    pyutilib.pyro.start_ns()

@pyomo_command('pyomo_nsc', "Execute the Pyro name server control tool for Pyomo")
def pyomo_nsc():
    pyutilib.pyro.start_nsc()

@pyomo_command('kill_pyro_mip_servers', "Terminate Pyomo's MIP solvers using Pyro")
def kill_pyro_mip_servers():
    if len(sys.argv) > 2:
        print("***Incorrect invocation - use: kill_pyro_mip_servers pid-filename")
        sys.exit(1)

    pid_filename = "pyro_mip_servers.pids"
    if len(sys.argv) == 2:
        pid_filename = sys.argv[1]

    print("Killing pyro mip servers specified in file="+pid_filename)

    pid_file = open(pid_filename, "r")
    for line in pid_file.readlines():
        pid = eval(string.strip(line))
        print("KILLING PID="+str(pid))
        os.kill(pid, signal.SIGTERM)
    pid_file.close()

@pyomo_command('launch_pyro_mip_servers', "Launch Pyomo's MIP solvers using Pyro")
def launch_pyro_mip_servers():
    if len(sys.argv) != 2:
        print("***Incorrect invocation - use: launch_pyro_mip_servers num-servers")
        sys.exit(1)

    num_servers = eval(sys.argv[1])

    print("Number of servers to launch="+str(num_servers))

    server_pids = []

    for i in range(1, num_servers+1):
        print("Launching server number "+str(i))
        output_filename = "pyro_mip_server"+str(i)+".out"
        # the "exec" ensures that (at least for bash) that the server process
        # will be the process returned, i.e., it becomes the child process - no
        # shell process intermediate. more correctly, exec exits the current
        # process before it does so (no fork).
        pid=subprocess.Popen("exec pyro_mip_server >& pyro_mip_server."+str(i)+".out", shell=True).pid
        server_pids.append(pid)

    # perhaps a better place would be in the users home directory, but I'll
    # worry about that a bit later.
    pid_output_filename = "pyro_mip_servers.pids"
    pid_output_file = open(pid_output_filename,"w")
    for pid in server_pids:
        pid_output_file.write(str(pid)+'\n')
    pid_output_file.close()

    print("PIDs for launched servers recorded in file="+pid_output_filename)

@pyomo_command('OSSolverService', "Launch an OS solver service")
def OSSolverService():
    import pyomo.opt

    if len(sys.argv) == 1:
        print("OSSolverService -osil <filename> -solver <name>")
        sys.exit(1)

    osilFile = None
    solver = None
    i=1
    while i<len(sys.argv):
        if sys.argv[i] == "-osil":
            i=i+1
            osilFile=sys.argv[i]
        elif sys.argv[i] == "-solver":
            i=i+1
            solver=sys.argv[i]
        i=i+1

    print("osilFile",osilFile,"solver",solver)

    with pyomo.opt.SolverFactory(solver) as opt:
        opt.solve(osilFile, rformat=pyomo.opt.ResultsFormat.osrl)

@pyomo_command('readsol', "Read and print a *.sol file")
def readsol():
    import pyomo.opt

    reader = pyomo.opt.ReaderFactory("sol")
    soln = reader(sys.argv[1])
    soln.write()

@pyomo_command('pyomo_python', "Launch script using Pyomo's python installation")
def pyomo_python(args=None):
    if args is None:
        args = sys.argv[1:]
    if args is None or len(args) == 0:
        console = code.InteractiveConsole()
        console.interact('Pyomo Python Console\n'+sys.version)
    else:
        cmd = sys.executable+' '+ ' '.join(args)
        pyutilib.subprocess.run(cmd, stdout=sys.stdout, stderr=sys.stderr)

@pyomo_command('pyomo', "The main command interface for Pyomo")
def pyomo(args=None):
    parser = pyomo.scripting.pyomo_parser.get_parser()
    if args is None:
        ret = parser.parse_args()
    else:
        ret = parser.parse_args(args)
    ret.func(ret)


@pyomo_command('results_schema', "Print the predefined schema for a results object")
def results_schema():
    if len(sys.argv) > 1:
        print("results_schema  - Print the predefined schema in a SolverResults object")
    options = Options(schema=True)
    r=SolverResults()
    repn = r._repn_(options)
    r.pprint(sys.stdout, options, repn=repn)




#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

from pyomo.environ import *

M = ConcreteModel()
M.x = Var(bounds=(0,1))
M.o = Objective(expr=2*M.x, sense=maximize)
M.c = Constraint(expr=M.x <= 0.5)

model = M




#! /usr/bin/env python
#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

#
# code to drive lagrgangeutils.py for a test
#

import sys
import operator
import traceback

import pyutilib.common

from pyomo.opt import SolverManagerFactory

from pyomo.pysp.scenariotree.instance_factory import \
   ScenarioTreeInstanceFactory
from pyomo.pysp.phinit import (construct_ph_options_parser,
                               GenerateScenarioTreeForPH,
                               PHFromScratch)
from pyomo.pysp.phutils import find_active_objective
from pyomo.pysp import lagrangeutils as lagrUtil

#########################################
def run(args=None):
##########################================================#########
   # to import plugins
   import pyomo.environ
   import pyomo.solvers.plugins.smanager.phpyro
   import pyomo.solvers.plugins.smanager.pyro

   def partialLagrangeParametric(args=None):
      print("lagrangeParam begins ")
      blanks = "                          "  # used for formatting print statements
      class Object(object): pass
      Result = Object()

# options used
      IndVarName = options.indicator_var_name
      CCStageNum = options.stage_num
      alphaTol = options.alpha_tol
      MaxMorePR = options.MaxMorePR # option to include up to this many PR points above F^* with all delta fixed
      outputFilePrefix = options.outputFilePrefix

# We write ScenarioList = name, probability
#          PRoptimal    = probability, min-cost, [selections]
#          PRmore       = probability, min-cost, [selections]
# ================ sorted by probability ========================
#
# These can be read to avoid re-computing points

      ph = PHFromScratch(options)
      Result.ph = ph
      rootnode = ph._scenario_tree._stages[0]._tree_nodes[0]   # use rootnode to loop over scenarios

      if find_active_objective(ph._scenario_tree._scenarios[0]._instance,safety_checks=True).is_minimizing():
         print("We are solving a MINIMIZATION problem.\n")
      else:
         print("We are solving a MAXIMIZATION problem.\n")

# initialize
      ScenarioList = []
      lambdaval = 0.
      lagrUtil.Set_ParmValue(ph,
                             options.lambda_parm_name,
                             lambdaval)

      # IMPORTANT: Preprocess the scenario instances
      #            before fixing variables, otherwise they
      #            will be preprocessed out of the expressions
      #            and the output_fixed_variable_bounds option
      #            will have no effect when we update the
      #            fixed variable values (and then assume we
      #            do not need to preprocess again because
      #            of this option).
      ph._preprocess_scenario_instances()

      lagrUtil.FixAllIndicatorVariables(ph, IndVarName, 0)
      for scenario in rootnode._scenarios:
         ScenarioList.append((scenario._name,
                              scenario._probability))

      # sorts from min to max probability
      ScenarioList.sort(key=operator.itemgetter(1))
      with open(outputFilePrefix+'ScenarioList.csv','w') as outFile:
         for scenario in ScenarioList:
            outFile.write(scenario[0]+ ", " +str(scenario[1])+"\n")
      Result.ScenarioList = ScenarioList

      print("lambda= "+str(lambdaval)+" ...run begins "+str(len(ScenarioList))+" scenarios")
      SolStat, zL = lagrUtil.solve_ph_code(ph, options)
      print("\t...ends")
      bL = Compute_ExpectationforVariable(ph,
                                          IndVarName,
                                          CCStageNum)
      if bL > 0:
         print("** bL = "+str(bL)+"  > 0")
         return Result

      print("Initial cost = "+str(zL)+"  for bL = "+str(bL))

      lagrUtil.FixAllIndicatorVariables(ph, IndVarName, 1)

      print("lambda= "+str(lambdaval)+" ...run begins")
      SolStat, zU = lagrUtil.solve_ph_code(ph, options)
      print("\t...ends")
      bU = Compute_ExpectationforVariable(ph,
                                          IndVarName,
                                          CCStageNum)
      if bU < 1:
            print("** bU = "+str(bU)+"  < 1")

      lagrUtil.FreeAllIndicatorVariables(ph, IndVarName)

      Result.lbz = [ [0,bL,zL], [None,bU,zU] ]
      Result.selections = [[], ScenarioList]
      NumIntervals = 1
      print("initial gap = "+str(1-zL/zU)+" \n")
      print("End of test; this is only a test.")

      return Result
################################
# LagrangeParametric ends here
################################

#### start run ####

   AllInOne = False

##########################
# options defined here
##########################
   try:
      conf_options_parser = construct_ph_options_parser("lagrange [options]")
      conf_options_parser.add_argument("--alpha",
                                     help="The alpha level for the chance constraint. Default is 0.05",
                                     action="store",
                                     dest="alpha",
                                     type=float,
                                     default=0.05)
      conf_options_parser.add_argument("--alpha-min",
                                     help="The min alpha level for the chance constraint. Default is None",
                                     action="store",
                                     dest="alpha_min",
                                     type=float,
                                     default=None)
      conf_options_parser.add_argument("--alpha-max",
                                     help="The alpha level for the chance constraint. Default is None",
                                     action="store",
                                     dest="alpha_max",
                                     type=float,
                                     default=None)
      conf_options_parser.add_argument("--min-prob",
                                     help="Tolerance for testing probability > 0. Default is 1e-5",
                                     action="store",
                                     dest="min_prob",
                                     type=float,
                                     default=1e-5)
      conf_options_parser.add_argument("--alpha-tol",
                                     help="Tolerance for testing equality to alpha. Default is 1e-5",
                                     action="store",
                                     dest="alpha_tol",
                                     type=float,
                                     default=1e-5)
      conf_options_parser.add_argument("--MaxMorePR",
                                     help="Generate up to this many additional PR points after response function. Default is 0",
                                     action="store",
                                     dest="MaxMorePR",
                                     type=int,
                                     default=0)
      conf_options_parser.add_argument("--outputFilePrefix",
                                     help="Output file name.  Default is ''",
                                     action="store",
                                     dest="outputFilePrefix",
                                     type=str,
                                     default="")
      conf_options_parser.add_argument("--stage-num",
                                     help="The stage number of the CC indicator variable (number, not name). Default is 2",
                                     action="store",
                                     dest="stage_num",
                                     type=int,
                                     default=2)
      conf_options_parser.add_argument("--lambda-parm-name",
                                     help="The name of the lambda parameter in the model. Default is lambdaMult",
                                     action="store",
                                     dest="lambda_parm_name",
                                     type=str,
                                     default="lambdaMult")
      conf_options_parser.add_argument("--indicator-var-name",
                                     help="The name of the indicator variable for the chance constraint. The default is delta",
                                     action="store",
                                     dest="indicator_var_name",
                                     type=str,
                                     default="delta")
      conf_options_parser.add_argument("--use-Loane-cuts",
                                     help="Add the Loane cuts if there is a gap. Default is False",
                                     action="store_true",
                                     dest="add_Loane_cuts",
                                     default=False)
      conf_options_parser.add_argument("--fofx-var-name",
                                     help="(Loane) The name of the model's auxiliary variable that is constrained to be f(x). Default is fofox",
                                     action="store",
                                     dest="fofx_var_name",
                                     type=str,
                                     default="fofx")
      conf_options_parser.add_argument("--solve-with-ph",
                                     help="Perform solves via PH rather than an EF solve. Default is False",
                                     action="store_true",
                                     dest="solve_with_ph",
                                     default=False)
      conf_options_parser.add_argument("--skip-graph",
                                     help="Do not show the graph at the end. Default is False (i.e. show the graph)",
                                     action="store_true",
                                     dest="skip_graph",
                                     default=False)
      conf_options_parser.add_argument("--write-xls",
                                     help="Write results into a xls file. Default is False",
                                     action="store_true",
                                     dest="write_xls",
                                     default=False)
      conf_options_parser.add_argument("--skip-ExpFlip",
                                     help="Do not show the results for flipping the indicator variable for each scenario. Default is False (i.e. show the flipping-results)",
                                     action="store_true",
                                     dest="skip_ExpFlip",
                                     default=False)
      conf_options_parser.add_argument("--HeurFlip",
                                     help="The number of solutions to evaluate after the heuristic. Default is 3. For 0 the heuristic flip gets skipped.",
                                     action="store",
                                     type=int,
                                     dest="HeurFlip",
                                     default=3)
      conf_options_parser.add_argument("--HeurMIP",
                                     help="The mipgap for the scenariowise solves in the heuristic. Default is 0.0001",
                                     action="store",
                                     type=float,
                                     dest="HeurMIP",
                                     default=0.0001)
      conf_options_parser.add_argument("--interactive",
                                     help="Enable interactive version of the code. Default is False.",
                                     action="store_true",
                                     dest="interactive",
                                     default=False)
      conf_options_parser.add_argument("--Lgap",
                                     help="The (relative) Lagrangian gap acceptable for the chance constraint. Default is 10^-4",
                                     action="store",
                                     type=float,
                                     dest="LagrangeGap",
                                     default=0.0001)
      conf_options_parser.add_argument("--lagrange-method",
                                     help="The Lagrange multiplier search method",
                                     action="store",
                                     dest="lagrange_search_method",
                                     type=str,
                                     default="tangential")
      conf_options_parser.add_argument("--max-lambda",
                                     help="The max value of the multiplier. Default=10^10",
                                     action="store",
                                     dest="max_lambda",
                                     type=float,
                                     default=10**10)
      conf_options_parser.add_argument("--min-lambda",
                                     help="The min value of the multiplier. Default=0.0",
                                     action="store",
                                     dest="min_lambda",
                                     type=float,
                                     default=0)
      conf_options_parser.add_argument("--min-probability",
                                     help="The min value of scenario probability. Default=10^-15",
                                     action="store",
                                     dest="min_probability",
                                     type=float,
                                     default=10**(-15))

################################################################

      options = conf_options_parser.parse_args(args=args)
      # temporary hack
      options._ef_options = conf_options_parser._ef_options
      options._ef_options.import_argparse(options)
   except SystemExit as _exc:
      # the parser throws a system exit if "-h" is specified - catch
      # it to exit gracefully.
      return _exc.code

   # load the reference model and create the scenario tree - no
   # scenario instances yet.
   if options.verbose:
      print("Loading reference model and scenario tree")
   #scenario_instance_factory, full_scenario_tree = load_models(options)
   scenario_instance_factory = \
        ScenarioTreeInstanceFactory(options.model_directory,
                                    options.instance_directory)

   full_scenario_tree = \
            GenerateScenarioTreeForPH(options,
                                      scenario_instance_factory)

   solver_manager = SolverManagerFactory(options.solver_manager_type)
   if solver_manager is None:
      raise ValueError("Failed to create solver manager of "
                       "type="+options.solver_manager_type+
                       " specified in call to PH constructor")
   if isinstance(solver_manager,
                 pyomo.solvers.plugins.smanager.phpyro.SolverManager_PHPyro):
      raise ValueError("PHPyro can not be used as the solver manager")

   try:

      if (scenario_instance_factory is None) or (full_scenario_tree is None):
         raise RuntimeError("***ERROR: Failed to initialize model and/or the scenario tree data.")

      # load_model gets called again, so lets make sure unarchived directories are used
      options.model_directory = scenario_instance_factory._model_filename
      options.instance_directory = scenario_instance_factory._scenario_tree_filename

      scenario_count = len(full_scenario_tree._stages[-1]._tree_nodes)

      # create ph objects for finding the solution. we do this even if
      # we're solving the extensive form

      if options.verbose:
         print("Loading scenario instances and initializing scenario tree for full problem.")

########## Here is where multiplier search is called ############
      Result = partialLagrangeParametric()
#####################################################################################

   finally:
      # delete temporary unarchived directories
      scenario_instance_factory.close()

   print("\nreturned from partialLagrangeParametric")

##########
def Compute_ExpectationforVariable(ph, IndVarName, CCStageNum):
   SumSoFar = 0.0
   node_probability = 0.0
   stage = ph._scenario_tree._stages[CCStageNum-1]
   for tree_node in stage._tree_nodes:
      for scenario in tree_node._scenarios:
         instance = ph._instances[scenario._name]
         #print "scenario._probability:",scenario._probability
         node_probability += scenario._probability
         #print "node_probability:",node_probability
         #print "getattr(instance, IndVarName).value:",getattr(instance, IndVarName).value
         SumSoFar += scenario._probability * getattr(instance, IndVarName).value
         #print "SumSoFar:",SumSoFar
   return SumSoFar / node_probability

#######################################

def Insert(newpoint,location,List):
    newList = []
    for i in range(location): newList.append(List[i])
    newList.append(newpoint)
    for i in range(location,len(List)): newList.append(List[i])
    return newList

#######################################

def ismember(List,member):  # designed to test 1st member of each list in List (ie, 1st column)
   for i in List:
      if len(i[0]) == 0: continue   # in case list contains empty list
      if i[0] == member: return True
   return

#######################################

def putcommas(num):
   snum = str(num)
   decimal = snum.find('.')
   if decimal >= 0:
      frac = snum[decimal:]
      snum = snum[0:decimal]
   else: frac = ''
   if len(snum) < 4: return snum + frac
   else: return putcommas(snum[:len(snum)-3]) + "," + snum[len(snum)-3:len(snum)] + frac

#######################################

def PrintPRpoints(PRlist):
   if len(PRlist) == 0:
      print("No PR points")
   else:
      print(str(len(PRlist))+" PR points:")
      blanks = "                      "
      print("            lambda        beta-probability       min cost ")
      for row in PRlist:
         b = round(row[1],4)
         z = round(row[2])
# lambda = row[0] could be float, string, or None
         sl = str(row[0])
         sl = blanks[0:20-len(sl)] + sl
         sb = str(b)
         sb = blanks[0:20-len(sb)] + sb
         sz = putcommas(z)
         sz = blanks[2:20-len(sz)] + sz
         print(sl+" "+sb+" "+sz)
      print("==================================================================\n")
   return

###########
def ZeroOneIndexListsforVariable(ph, IndVarName, CCStageNum):
   # return lists across scenarios of the zero value scenarios and one value scenarios
   # for unindexed variable in the ph object for a stage (one based)
   # in this routine we trust that it is binary
   ZerosList = []
   OnesList = []

   stage = ph._scenario_tree._stages[CCStageNum-1]
   for tree_node in stage._tree_nodes:
      for scenario in tree_node._scenarios:
         instance = ph._instances[scenario._name]
         locval = getattr(instance, IndVarName).value
         #print locval
         if locval < 0.5:
            ZerosList.append(scenario)
         else:
            OnesList.append(scenario)
   return [ZerosList,OnesList]

###########
def PrintanIndexList(IndList):
   # show some useful information about an index list (note: indexes are scenarios)
   print("Zeros:")
   for i in IndList[0]:
      print(i._name)
   print("Ones:")
   for i in IndList[1]:
      print(i._name)

###########
def ReturnIndexListNames(IndList):
   ListNames=[[],[]]
   #print "Zeros:"
   for i in IndList[0]:
      ListNames[0].append(i._name)
   #print "Ones:"
   for i in IndList[1]:
      ListNames[1].append(i._name)
   return ListNames

#
# the main script routine starts here
#

def main(args=None):

   try:
      run(args)
   except ValueError:
      msg = sys.exc_info()[1]
      print("VALUE ERROR:")
      print(msg)
   except IOError:
      msg = sys.exc_info()[1]
      print("IO ERROR:")
      print(msg)
   except pyutilib.common.ApplicationError:
      msg = sys.exc_info()[1]
      print("APPLICATION ERROR:")
      print(msg)
   except RuntimeError:
      msg = sys.exc_info()[1]
      print("RUN-TIME ERROR:")
      print(msg)
   except:
      print("Encountered unhandled exception"+str(sys.exc_info()[0]))
      traceback.print_exc()

if __name__ == "__main__":

   run()
